/* Gtk guitar effects processor ver. 0.2.1 */
/* Copyright (c) 2000 Wojciech Rokita (cjrocky@polbox.com) */
#include <config.h>
#include <gtk/gtk.h>
#include "plugin.h"
#include "oss.h"
#include <dirent.h>
#include <sys/stat.h>
#include <dlfcn.h>
#include <string.h>
#include <memory.h>
#include <sched.h>
#include <pthread.h>


#ifndef PLUGINS_PATH
#define PLUGINS_PATH "/usr/lib64/" PACKAGE "-" VERSION "/plugins" 
#endif

#define MAX_NUMBER_OF_PLUGINS 32

typedef struct
{
	char *name;
	char comment[80];
	void *handle;
	void *data;
	void (*cleanup)(void *data);
	void *(*init)(void);
	void (*configure)(void *data);
 	int (*process)(void *data, short int *smpdata, int size, int sampling_frequency, int nchan);
} InternalPlugin;


void quit_rtdsp_menu(GtkWidget *, gpointer);
void configure_devices(GtkWidget *, gpointer);
void about(GtkWidget *, gpointer);

InternalPlugin loaded_plugins_data[MAX_NUMBER_OF_PLUGINS];

int loaded_plugins_num=0;
int number_of_founded_plugins=0;
int selected_plugin_num=0;
int selected_plugin_add_num=0;
int is_playing=0;
GtkWidget *quit_window=NULL;
GtkWidget *add_window=NULL;
GtkWidget *plugin_list;
GtkWidget *entry_comment;
GtkWidget *device_config_window;
GtkWidget *about_window;
GtkItemFactoryEntry menu_items[]=
{{"/_File", NULL, NULL, 0, "<Branch>"},
 {"/File/_Quit", NULL, quit_rtdsp_menu, 0, NULL},
 {"/_Options", NULL, NULL, 0, "<Branch>"},
 {"/Options/Device", NULL, configure_devices, 0, NULL},
 {"/_Help", NULL, NULL, 0, "<LastBranch>"},
 {"/Help/About", NULL, about, 0, NULL}
};

char about_text[]="Gtk Guitar Effects Processor ver. " VERSION;


struct
{
	int sampling_freq;
	int buffer_size;
	char *snd_dev_name;
	int nchan;
} temporary_config_entry;


pthread_t audio_thread;



void about_destroy(GtkWidget *w, gpointer d)
{
    if(about_window != NULL) gtk_widget_destroy(about_window);
    about_window=NULL;
}

void about(GtkWidget *w, gpointer d)
{
    GtkWidget *text, *button, *vbox;
	if(about_window) return;
	about_window=gtk_window_new(GTK_WINDOW_TOPLEVEL);
	gtk_window_set_title(GTK_WINDOW(about_window), "About");
	gtk_container_set_border_width(GTK_CONTAINER(about_window), 5);
	gtk_signal_connect(GTK_OBJECT(about_window), "destroy", GTK_SIGNAL_FUNC(about_destroy), NULL);
	vbox=gtk_vbox_new(FALSE, 0);
	gtk_container_add(GTK_CONTAINER(about_window), vbox);
	gtk_widget_show(vbox);
	text=gtk_text_new(NULL, NULL);
	gtk_text_set_editable(GTK_TEXT(text), FALSE);
	gtk_text_insert(GTK_TEXT(text), NULL, NULL, NULL, about_text, strlen(about_text));
	gtk_box_pack_start(GTK_BOX(vbox), text, TRUE, TRUE, 0);
	gtk_widget_show(text);
	button=gtk_button_new_with_label("Ok");	
	gtk_box_pack_start(GTK_BOX(vbox), button, FALSE, FALSE, 0);
	gtk_signal_connect(GTK_OBJECT(button), "clicked", GTK_SIGNAL_FUNC(about_destroy), NULL);
	gtk_widget_show(button);
	gtk_widget_show(about_window);
}

void add_window_close(GtkWidget *button, gpointer data)
{
	gtk_widget_destroy(add_window);
	add_window=NULL;
}



int load_plugin_from_disk(char *filename, InternalPlugin *where)
{
	char fullfilename[256]=PLUGINS_PATH;
	Plugin *plugininfo;
	Plugin *(*get_plugin_info)(void);
	void *handle;
	

	strcat(fullfilename, "/");
	strcat(fullfilename, filename);
	if((handle=dlopen(fullfilename, RTLD_NOW)) == NULL) return 1;
	if((get_plugin_info=dlsym(handle, "get_plugin_info")) != NULL)
	{
		where->handle=handle;
		plugininfo=get_plugin_info();
		where->name=plugininfo->name;
		where->init=plugininfo->init;
		where->cleanup=plugininfo->cleanup;
		where->process=plugininfo->process;
		where->configure=plugininfo->configure;
		where->data=NULL;
		return 0;
	}	
	else
	{
		dlclose(handle);
		return 1;
	}
}



void add_window_append(GtkWidget *button, gpointer data)
{
	gchar *rownames[2];
	int new_plugin_list_row;
	int rownum;
	InternalPlugin temp_plugin;

	if(loaded_plugins_num >= MAX_NUMBER_OF_PLUGINS) return;
	if(number_of_founded_plugins == 0) return;

	gtk_clist_get_text((GtkCList *)data, selected_plugin_add_num, 0, &rownames[0]);
	if(load_plugin_from_disk(rownames[0], &temp_plugin))	
	{
		return;
	}

	if(loaded_plugins_num > 0)
	{
		new_plugin_list_row=selected_plugin_num + 1;
		if(selected_plugin_num < (loaded_plugins_num - 1))
		{
			for(rownum=(loaded_plugins_num - 1); rownum > selected_plugin_num; rownum--)
			{
				memcpy(&loaded_plugins_data[rownum + 1],&loaded_plugins_data[rownum], sizeof(InternalPlugin));
			}
		}
	}
	else
	{
		new_plugin_list_row=0;
	}

	memcpy(&loaded_plugins_data[new_plugin_list_row], &temp_plugin, sizeof(InternalPlugin));
	gtk_clist_get_text((GtkCList *)data, selected_plugin_add_num, 0, &rownames[1]);
	gtk_clist_get_text((GtkCList *)data, selected_plugin_add_num, 1, &rownames[0]);

	rownames[1]=gtk_entry_get_text(GTK_ENTRY(entry_comment));
	gtk_clist_insert(GTK_CLIST(plugin_list), new_plugin_list_row, rownames);
	if(loaded_plugins_num == 0) gtk_clist_select_row(GTK_CLIST(plugin_list), 0, 0);
	else
	{
		selected_plugin_num++;
		gtk_clist_select_row(GTK_CLIST(plugin_list), selected_plugin_num, 0);
	}

	if(loaded_plugins_data[new_plugin_list_row].init != NULL)
	{
		loaded_plugins_data[new_plugin_list_row].data=loaded_plugins_data[new_plugin_list_row].init();
	}
	
	loaded_plugins_num++;
}




void add_window_insert(GtkWidget *button, gpointer data)
{
	gchar *rownames[2];
	int new_plugin_list_row;
	int rownum;
	InternalPlugin temp_plugin;

	if(loaded_plugins_num >= MAX_NUMBER_OF_PLUGINS) return;
	if(number_of_founded_plugins == 0) return;

	gtk_clist_get_text((GtkCList *)data, selected_plugin_add_num, 0, &rownames[0]);
	if(load_plugin_from_disk(rownames[0], &temp_plugin))	
	{
		return;
	}

	if(loaded_plugins_num > 0)
	{
		new_plugin_list_row=selected_plugin_num;
		if(new_plugin_list_row < 0) new_plugin_list_row=0;
		for(rownum=(loaded_plugins_num - 1); rownum >= selected_plugin_num; rownum--)
		{
			memcpy(&loaded_plugins_data[rownum + 1],&loaded_plugins_data[rownum], sizeof(InternalPlugin));
		}
	}
	else
	{
		new_plugin_list_row=0;
	}

	memcpy(&loaded_plugins_data[new_plugin_list_row], &temp_plugin, sizeof(InternalPlugin));
	gtk_clist_get_text((GtkCList *)data, selected_plugin_add_num, 0, &rownames[1]);
	gtk_clist_get_text((GtkCList *)data, selected_plugin_add_num, 1, &rownames[0]);

	rownames[1]=gtk_entry_get_text(GTK_ENTRY(entry_comment));
	gtk_clist_insert(GTK_CLIST(plugin_list), new_plugin_list_row, rownames);
	gtk_clist_select_row(GTK_CLIST(plugin_list), new_plugin_list_row + 1, 0);
	if(loaded_plugins_num == 0) gtk_clist_select_row(GTK_CLIST(plugin_list), 0, 0);

	if(loaded_plugins_data[new_plugin_list_row].init != NULL)
	{
		loaded_plugins_data[new_plugin_list_row].data=loaded_plugins_data[new_plugin_list_row].init();
	}
	
	loaded_plugins_num++;
}









gchar *get_plugin_name(gchar *filename)
{
	gchar fullfilename[256]=PLUGINS_PATH;
	static gchar pname[256];
	void *handle;
	Plugin *plugininfo;
	Plugin *(*get_plugin_info)(void);

	if(strcmp(&filename[strlen(filename) - 2], "so")) return NULL;
	strcat(fullfilename, "/");
	strcat(fullfilename, filename);
	if((handle=dlopen(fullfilename, RTLD_NOW)) == NULL) return NULL;
	if((get_plugin_info=dlsym(handle, "get_plugin_info")) != NULL)
	{
    	plugininfo=get_plugin_info();
		strcpy(pname, plugininfo->name);
		dlclose(handle);
		return pname;
	}
	
	dlclose(handle);
	return NULL;
}

void find_all_plugins(GtkWidget *list)
{
	DIR *dp;
	struct dirent *wpis;
	struct stat statbufor;
	gchar *list_row[2];

	number_of_founded_plugins=0;
	if((dp=opendir(PLUGINS_PATH)) == NULL) return;
	chdir(PLUGINS_PATH);

	while((wpis=readdir(dp)) != NULL)
	{
		stat(wpis->d_name, &statbufor);
		if(S_ISREG(statbufor.st_mode))
		{
			list_row[1]=get_plugin_name(wpis->d_name);
			if(list_row[1] != NULL)
			{
				list_row[0]=wpis->d_name;
				gtk_clist_append(GTK_CLIST(list), list_row);
				number_of_founded_plugins++;
			}
		}
	}

	closedir(dp);
}


void select_row_add_list(GtkWidget *list, gint row, gint column, GdkEventButton *event, gpointer data)
{
	selected_plugin_add_num=row;
}


void add_plugin_to_list(GtkWidget *button, gpointer data)
{
	GtkWidget *vbox;
	GtkWidget *hbox;
	GtkWidget *button_box;
	GtkWidget *button_append;
	GtkWidget *button_insert;
	GtkWidget *button_cancel;
	GtkWidget *scrolled_win;
	GtkWidget *list;
	GtkWidget *entry_frame;

	if(add_window != NULL) return;
	add_window=gtk_window_new(GTK_WINDOW_DIALOG);
	gtk_container_set_border_width(GTK_CONTAINER(add_window), 10);
	gtk_window_set_title(GTK_WINDOW(add_window), "Add plugin");	
	gtk_widget_set_usize(add_window, 350, 300);
	gtk_signal_connect(GTK_OBJECT(add_window), "destroy", GTK_SIGNAL_FUNC(add_window_close), NULL);

	vbox=gtk_vbox_new(FALSE, 0);
	hbox=gtk_hbox_new(FALSE, 0);
	button_box=gtk_vbox_new(FALSE, 0);
	button_append=gtk_button_new_with_label("Append");
	button_insert=gtk_button_new_with_label("Insert");
	button_cancel=gtk_button_new_with_label("Cancel");
	gtk_signal_connect(GTK_OBJECT(button_cancel), "clicked", GTK_SIGNAL_FUNC(add_window_close), NULL);

	list=gtk_clist_new(2);
	gtk_signal_connect(GTK_OBJECT(button_append), "clicked", GTK_SIGNAL_FUNC(add_window_append), (gpointer)list);
	gtk_signal_connect(GTK_OBJECT(button_insert), "clicked", GTK_SIGNAL_FUNC(add_window_insert), (gpointer)list);
	gtk_clist_set_column_title(GTK_CLIST(list), 0, "Plugin files");
	gtk_clist_set_column_title(GTK_CLIST(list), 1, "Plugin names");
	gtk_clist_column_titles_show(GTK_CLIST(list));
	gtk_clist_column_titles_passive(GTK_CLIST(list));
	gtk_clist_set_selection_mode(GTK_CLIST(list), GTK_SELECTION_BROWSE);
	gtk_clist_set_shadow_type(GTK_CLIST(list), GTK_SHADOW_OUT);
	gtk_clist_set_column_width(GTK_CLIST(list), 0, 100);
	gtk_clist_set_column_width(GTK_CLIST(list), 1, 100);
	gtk_signal_connect(GTK_OBJECT(list), "select_row", GTK_SIGNAL_FUNC(select_row_add_list), NULL);
	
	find_all_plugins(list);
	if(number_of_founded_plugins > 0) gtk_clist_select_row(GTK_CLIST(list), 0, 0);

	scrolled_win=gtk_scrolled_window_new(NULL, NULL);
	gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(scrolled_win), GTK_POLICY_AUTOMATIC, GTK_POLICY_AUTOMATIC);
	gtk_scrolled_window_add_with_viewport(GTK_SCROLLED_WINDOW(scrolled_win), list);

	gtk_box_pack_start(GTK_BOX(hbox), scrolled_win, TRUE, TRUE, 0);
	gtk_box_pack_start(GTK_BOX(hbox), button_box, FALSE, FALSE, 10);
	gtk_box_pack_start(GTK_BOX(button_box), button_append, FALSE, FALSE, 0);
	gtk_box_pack_start(GTK_BOX(button_box), button_insert, FALSE, FALSE, 0);
	gtk_box_pack_start(GTK_BOX(button_box), button_cancel, FALSE, FALSE, 0);
	
	gtk_box_pack_start(GTK_BOX(vbox), hbox, TRUE, TRUE, 0);

	entry_frame=gtk_frame_new("Plugin's comment");
	gtk_box_pack_start(GTK_BOX(vbox), entry_frame, FALSE, FALSE, 10);	

	entry_comment=gtk_entry_new_with_max_length(79);
	gtk_container_add(GTK_CONTAINER(entry_frame), entry_comment);

	gtk_container_add(GTK_CONTAINER(add_window), vbox);

	gtk_widget_show(button_append);
	gtk_widget_show(button_insert);
	gtk_widget_show(button_cancel);
	gtk_widget_show(button_box);
	gtk_widget_show(list);
	gtk_widget_show(scrolled_win);
	gtk_widget_show(entry_comment);
	gtk_widget_show(entry_frame);
	gtk_widget_show(hbox);
	gtk_widget_show(vbox);
	gtk_widget_show(add_window);
}

void remove_plugin_from_list(GtkWidget *button, gpointer data)
{
	gchar *rownames[2];
	int rownum;

	if(loaded_plugins_num == 0) return;
	if(loaded_plugins_data[selected_plugin_num].cleanup != NULL)
	{
		loaded_plugins_data[selected_plugin_num].cleanup(loaded_plugins_data[selected_plugin_num].data);
	}
	dlclose(loaded_plugins_data[selected_plugin_num].handle);
	if(selected_plugin_num < (loaded_plugins_num - 1))
	{
		for(rownum=selected_plugin_num; rownum < loaded_plugins_num; rownum++)
		{
			memcpy(&loaded_plugins_data[rownum], &loaded_plugins_data[rownum + 1], sizeof(InternalPlugin));		
		}
	}

	loaded_plugins_num--;
	memset(&loaded_plugins_data[loaded_plugins_num], 0, sizeof(InternalPlugin));
	gtk_clist_remove(GTK_CLIST(plugin_list), selected_plugin_num);
	if(selected_plugin_num == loaded_plugins_num)
	    gtk_clist_select_row(GTK_CLIST(plugin_list), selected_plugin_num - 1, 0);
	else
	    gtk_clist_select_row(GTK_CLIST(plugin_list), selected_plugin_num, 0);
}


void configure_devices_freq_11(GtkWidget *but, gpointer data)
{
	temporary_config_entry.sampling_freq=11025;
}

void configure_devices_freq_22(GtkWidget *but, gpointer data)
{
	temporary_config_entry.sampling_freq=22050;
}

void configure_devices_freq_44(GtkWidget *but, gpointer data)
{
	temporary_config_entry.sampling_freq=44100;
}



void configure_devices_buf_64(GtkWidget *but, gpointer data)
{
	temporary_config_entry.buffer_size=64;
}

void configure_devices_buf_128(GtkWidget *but, gpointer data)
{
	temporary_config_entry.buffer_size=128;
}


void configure_devices_buf_256(GtkWidget *but, gpointer data)
{
	temporary_config_entry.buffer_size=256;
}

void configure_devices_buf_1024(GtkWidget *but, gpointer data)
{
	temporary_config_entry.buffer_size=1024;
}

void configure_devices_buf_2048(GtkWidget *but, gpointer data)
{
	temporary_config_entry.buffer_size=2048;
}

void configure_devices_stereo(GtkWidget *but, gpointer data)
{
	temporary_config_entry.nchan=2;
}

void configure_devices_mono(GtkWidget *but, gpointer data)
{
	temporary_config_entry.nchan=1;
}


void configure_devices_ok(GtkWidget *w, gpointer data)
{
	int plgnnum;

	temporary_config_entry.snd_dev_name=gtk_entry_get_text(GTK_ENTRY(data));

	strcpy(sound_device_name, temporary_config_entry.snd_dev_name);
	sampling_frequency=temporary_config_entry.sampling_freq;
	blocksize=temporary_config_entry.buffer_size;
	nchan=temporary_config_entry.nchan;	
}


void configure_devices_cancel(GtkWidget *w, gpointer data)
{
	gtk_widget_destroy((GtkWidget *)data);
	device_config_window=NULL;
}


void configure_devices(GtkWidget *w, gpointer d)
{
	GtkWidget *vbox;
	GtkWidget *device_name_entry;
	GtkWidget *device_name_frame;
	GtkWidget *button_box;
	GtkWidget *sampling_freq_frame;
	GtkWidget *sampling_freq_button;
	GSList    *sampling_freq_button_group;
	GtkWidget *buffer_size_frame;
	GtkWidget *buffer_size_button;
	GSList    *buffer_size_button_group;
	GtkWidget *button_ok;
	GtkWidget *button_cancel;
	GtkWidget *button_box2;
	GtkWidget *stereo_frame;
	GtkWidget *stereo_button;
	GSList	  *stereo_button_group;


    if(device_config_window != NULL) return;
    if(is_playing) return;
	
	device_config_window=gtk_window_new(GTK_WINDOW_DIALOG);
	gtk_container_set_border_width(GTK_CONTAINER(device_config_window), 10);
	vbox=gtk_vbox_new(FALSE, 0);	
	gtk_container_add(GTK_CONTAINER(device_config_window), vbox);
	
	device_name_frame=gtk_frame_new("Input/Output device");
	gtk_frame_set_shadow_type(GTK_FRAME(device_name_frame), GTK_SHADOW_OUT);
	gtk_widget_show(device_name_frame);

	device_name_entry=gtk_entry_new_with_max_length(30);
	gtk_entry_set_text(GTK_ENTRY(device_name_entry), sound_device_name);
	gtk_container_add(GTK_CONTAINER(device_name_frame), device_name_entry);	
	gtk_widget_show(device_name_entry);

	sampling_freq_frame=gtk_frame_new("Sampling frequency");
	gtk_frame_set_shadow_type(GTK_FRAME(sampling_freq_frame), GTK_SHADOW_OUT);
	gtk_widget_show(sampling_freq_frame);
	
	buffer_size_frame=gtk_frame_new("Buffer size");
	gtk_frame_set_shadow_type(GTK_FRAME(buffer_size_frame), GTK_SHADOW_OUT);
	gtk_widget_show(buffer_size_frame);
	
	stereo_frame=gtk_frame_new("Mono/Stereo");
	gtk_frame_set_shadow_type(GTK_FRAME(stereo_frame), GTK_SHADOW_OUT);
	gtk_widget_show(stereo_frame);


	button_box=gtk_vbox_new(FALSE, 0);
	gtk_container_add(GTK_CONTAINER(sampling_freq_frame), button_box);
	gtk_widget_show(button_box);

/*	sampling_freq_button=gtk_radio_button_new_with_label(NULL, "11 kHz");
	gtk_box_pack_start(GTK_BOX(button_box), sampling_freq_button, FALSE, FALSE, 0);
	gtk_widget_show(sampling_freq_button);
	gtk_signal_connect(GTK_OBJECT(sampling_freq_button), "clicked", GTK_SIGNAL_FUNC(configure_devices_freq_11), NULL);
	if(sampling_frequency == 11025)
		gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (sampling_freq_button), TRUE);
	sampling_freq_button_group=gtk_radio_button_group(GTK_RADIO_BUTTON(sampling_freq_button));*/

	sampling_freq_button=gtk_radio_button_new_with_label(NULL, "22 kHz");
	gtk_signal_connect(GTK_OBJECT(sampling_freq_button), "clicked", GTK_SIGNAL_FUNC(configure_devices_freq_22), NULL);
	gtk_box_pack_start(GTK_BOX(button_box), sampling_freq_button, FALSE, FALSE, 0);
	gtk_widget_show(sampling_freq_button);
	if(sampling_frequency == 22050)
		gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (sampling_freq_button), TRUE);
	sampling_freq_button_group=gtk_radio_button_group(GTK_RADIO_BUTTON(sampling_freq_button));

	sampling_freq_button=gtk_radio_button_new_with_label(sampling_freq_button_group, "44 kHz");
	gtk_signal_connect(GTK_OBJECT(sampling_freq_button), "clicked", GTK_SIGNAL_FUNC(configure_devices_freq_44), NULL);
	gtk_box_pack_start(GTK_BOX(button_box), sampling_freq_button, FALSE, FALSE, 0);
	gtk_widget_show(sampling_freq_button);
	if(sampling_frequency == 44100)
		gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (sampling_freq_button), TRUE);



	button_box=gtk_vbox_new(FALSE, 0);
	gtk_container_add(GTK_CONTAINER(buffer_size_frame), button_box);
	gtk_widget_show(button_box);



	buffer_size_button=gtk_radio_button_new_with_label(NULL, "64 bytes");
	gtk_signal_connect(GTK_OBJECT(buffer_size_button), "clicked", GTK_SIGNAL_FUNC(configure_devices_buf_64), NULL);
	gtk_box_pack_start(GTK_BOX(button_box), buffer_size_button, FALSE, FALSE, 0);
	gtk_widget_show(buffer_size_button);
	if(blocksize == 64)
		gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (buffer_size_button), TRUE);

	buffer_size_button_group=gtk_radio_button_group(GTK_RADIO_BUTTON(buffer_size_button));
	buffer_size_button=gtk_radio_button_new_with_label(buffer_size_button_group, "128 bytes");
	gtk_signal_connect(GTK_OBJECT(buffer_size_button), "clicked", GTK_SIGNAL_FUNC(configure_devices_buf_128), NULL);
	if(blocksize == 128)
		gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (buffer_size_button), TRUE);
	gtk_box_pack_start(GTK_BOX(button_box), buffer_size_button, FALSE, FALSE, 0);
	gtk_widget_show(buffer_size_button);

	buffer_size_button_group=gtk_radio_button_group(GTK_RADIO_BUTTON(buffer_size_button));
	buffer_size_button=gtk_radio_button_new_with_label(buffer_size_button_group, "256 bytes");
	gtk_signal_connect(GTK_OBJECT(buffer_size_button), "clicked", GTK_SIGNAL_FUNC(configure_devices_buf_256), NULL);
	if(blocksize == 256)
		gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (buffer_size_button), TRUE);
	gtk_box_pack_start(GTK_BOX(button_box), buffer_size_button, FALSE, FALSE, 0);
	gtk_widget_show(buffer_size_button);

	buffer_size_button_group=gtk_radio_button_group(GTK_RADIO_BUTTON(buffer_size_button));
	buffer_size_button=gtk_radio_button_new_with_label(buffer_size_button_group, "1024 bytes");
	gtk_signal_connect(GTK_OBJECT(buffer_size_button), "clicked", GTK_SIGNAL_FUNC(configure_devices_buf_1024), NULL);
	if(blocksize == 1024)
		gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (buffer_size_button), TRUE);
	gtk_box_pack_start(GTK_BOX(button_box), buffer_size_button, FALSE, FALSE, 0);
	gtk_widget_show(buffer_size_button);

	buffer_size_button_group=gtk_radio_button_group(GTK_RADIO_BUTTON(buffer_size_button));
	buffer_size_button=gtk_radio_button_new_with_label(buffer_size_button_group, "2048 bytes");
	gtk_signal_connect(GTK_OBJECT(buffer_size_button), "clicked", GTK_SIGNAL_FUNC(configure_devices_buf_2048), NULL);
	gtk_box_pack_start(GTK_BOX(button_box), buffer_size_button, FALSE, FALSE, 0);
	gtk_widget_show(buffer_size_button);
	if(blocksize == 2048)
		gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (buffer_size_button), TRUE);

	button_box2=gtk_hbox_new(TRUE, 0);
	button_ok=gtk_button_new_with_label("Ok");
	gtk_signal_connect(GTK_OBJECT(button_ok), "clicked", GTK_SIGNAL_FUNC(configure_devices_ok), device_name_entry);
	gtk_signal_connect(GTK_OBJECT(button_ok), "clicked", GTK_SIGNAL_FUNC(configure_devices_cancel), device_config_window);
	button_cancel=gtk_button_new_with_label("Cancel");
	gtk_signal_connect(GTK_OBJECT(button_cancel), "clicked", GTK_SIGNAL_FUNC(configure_devices_cancel), device_config_window);
	gtk_widget_show(button_ok);
	gtk_widget_show(button_cancel);
	gtk_box_pack_start(GTK_BOX(button_box2), button_ok, FALSE, TRUE, 10);
	gtk_box_pack_start(GTK_BOX(button_box2), button_cancel, FALSE, TRUE, 10);
	gtk_widget_show(button_box2);
	
	
	button_box=gtk_vbox_new(FALSE, 0);
	gtk_container_add(GTK_CONTAINER(stereo_frame), button_box);
	gtk_widget_show(button_box);

	stereo_button=gtk_radio_button_new_with_label(NULL, "Mono");
	gtk_signal_connect(GTK_OBJECT(stereo_button), "clicked", GTK_SIGNAL_FUNC(configure_devices_mono), NULL);
	gtk_box_pack_start(GTK_BOX(button_box), stereo_button, FALSE, FALSE, 0);
	gtk_widget_show(stereo_button);
	if(nchan == 1)
		gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (stereo_button), TRUE);

	stereo_button_group=gtk_radio_button_group(GTK_RADIO_BUTTON(stereo_button));
	stereo_button=gtk_radio_button_new_with_label(stereo_button_group, "Stereo");
	gtk_signal_connect(GTK_OBJECT(stereo_button), "clicked", GTK_SIGNAL_FUNC(configure_devices_stereo), NULL);
	if(nchan == 2)
		gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (stereo_button), TRUE);
	gtk_box_pack_start(GTK_BOX(button_box), stereo_button, FALSE, FALSE, 0);
	gtk_widget_show(stereo_button);

	gtk_box_pack_start(GTK_BOX(vbox), device_name_frame, FALSE, FALSE, 0);
	gtk_box_pack_start(GTK_BOX(vbox), sampling_freq_frame, FALSE, FALSE, 0);
	gtk_box_pack_start(GTK_BOX(vbox), buffer_size_frame, FALSE, FALSE, 0);
	gtk_box_pack_start(GTK_BOX(vbox), stereo_frame, FALSE, FALSE, 0);	
	gtk_box_pack_start(GTK_BOX(vbox), button_box2, FALSE, FALSE, 10);	
	gtk_widget_show(vbox);
	gtk_widget_show(device_config_window);
}



void configure_plugin(GtkWidget *button, gpointer data)
{
	if(loaded_plugins_data[selected_plugin_num].configure == NULL) return;
	loaded_plugins_data[selected_plugin_num].configure(loaded_plugins_data[selected_plugin_num].data);
}


void print_error_message_destroy(GtkWidget *w, gpointer data)
{
	gtk_widget_destroy((GtkWidget *)data);
}



void print_error_message(gchar *message)
{
    GtkWidget *window;
	GtkWidget *text;
	GtkWidget *button_ok;
	GtkWidget *vbox;

	window=gtk_window_new(GTK_WINDOW_DIALOG);
	gtk_container_set_border_width(GTK_CONTAINER(window), 10);	
	gtk_window_set_title(GTK_WINDOW(window), "Error");
	gtk_widget_set_usize(window, 200, 100);
	gtk_signal_connect(GTK_OBJECT(window), "destroy", GTK_SIGNAL_FUNC(print_error_message_destroy), window);
	
	vbox=gtk_vbox_new(TRUE, 10);
	gtk_widget_show(vbox);

	button_ok=gtk_button_new_with_label("Ok");		
	gtk_widget_show(button_ok);
	gtk_signal_connect(GTK_OBJECT(button_ok), "clicked", GTK_SIGNAL_FUNC(print_error_message_destroy), window);

	text=gtk_label_new(message);
	gtk_widget_show(text);

	gtk_box_pack_start(GTK_BOX(vbox), text, FALSE, FALSE, 0);
	gtk_box_pack_start(GTK_BOX(vbox), button_ok, FALSE, FALSE, 0);

	gtk_container_add(GTK_CONTAINER(window), vbox);
	gtk_widget_show(window);
}



void *audio_thread_start(void *abc)
{
    int plgnnum;
    
    while(is_playing)
    {
	sound_read();

	if(loaded_plugins_num > 0)
	    for(plgnnum=0; plgnnum < loaded_plugins_num; plgnnum++)
	    {
		if(loaded_plugins_data[plgnnum].process != NULL)
		{
		    loaded_plugins_data[plgnnum].process(loaded_plugins_data[plgnnum].data,
							 process_buffer, blocksize, sampling_frequency, nchan);
		}
	    }

	sound_write();
    }

    return NULL;
}


void process_start_stop(GtkWidget *button, gpointer data)
{
	int plgnnum;
	int error;
	int max_priority;
	struct sched_param param;
	
	if(is_playing)
	{
		is_playing=0;
		gtk_label_set((GtkLabel *)data , "Start");
		
		pthread_join(audio_thread, NULL);
		
		sound_stop_playing();
	}
	else
	{
		if((error=sound_start_playing()) != 0)
		{
			switch(error)
			{
				case 1:       /* Device not available or busy */
					print_error_message("Device is not available or busy");
					break;
				case 2:       /* Device doesn't support full duplex */
					print_error_message("Device doesn't support full-duplex mode");
					break;
				case 3:       /* Device doesn't support such sampling frequency */
					print_error_message("Device doesn't support such sampling frequency");
					break;
			}

			return;
		}

		is_playing=1;
		gtk_label_set((GtkLabel *)data, "Stop");

	        max_priority = sched_get_priority_max(SCHED_FIFO);
	        param.sched_priority = max_priority;

		/* trying to set highest priority */
	        sched_setscheduler(0, SCHED_FIFO, &param);

		if (pthread_create(&audio_thread, NULL, audio_thread_start, NULL)) 
		{
		    printf("Can't create audio thread.");
		}
	}
}


void select_row_plugin_list(GtkWidget *list, gint row, gint column, GdkEventButton *event, gpointer data)
{
	selected_plugin_num=row;
}



void quit_rtdsp_menu(GtkWidget *window, gpointer data)
{
	quit_rtdsp(NULL, NULL, NULL);
}


void quit_rtdsp_ok(GtkWidget *window, gpointer data)
{
	if(is_playing)
	{
		is_playing=0;
		sound_stop_playing();
	}
	gtk_main_quit();
}


void quit_rtdsp_cancel(GtkWidget *window, gpointer data)
{
	gtk_widget_destroy(quit_window);
	quit_window=NULL;
}


gint quit_rtdsp(GtkWidget *window, GdkEvent *event, gpointer data)
{
	GtkWidget *button_ok;
	GtkWidget *button_cancel;
	GtkWidget *button_box;
	GtkWidget *vbox;
	GtkWidget *question;
	
	if(quit_window!=NULL) return(TRUE);
	quit_window=gtk_window_new(GTK_WINDOW_DIALOG);
	gtk_window_set_title(GTK_WINDOW(quit_window), "Quit?");
	gtk_container_set_border_width(GTK_CONTAINER(quit_window), 20);
	button_ok=gtk_button_new_with_label("Ok");
	button_cancel=gtk_button_new_with_label("Cancel");
	button_box=gtk_hbox_new(TRUE, 10);
	vbox=gtk_vbox_new(TRUE, 10);
	question=gtk_label_new("ARE YOU SURE YOU WANT TO QUIT?");
	gtk_box_pack_start(GTK_BOX(vbox), question, FALSE, FALSE, 0);
	gtk_box_pack_start(GTK_BOX(vbox), button_box, FALSE, FALSE, 0);
	gtk_widget_show(vbox);
	gtk_widget_show(question);

	gtk_signal_connect(GTK_OBJECT(button_ok), "clicked", GTK_SIGNAL_FUNC(quit_rtdsp_ok), NULL);
	gtk_signal_connect(GTK_OBJECT(button_cancel), "clicked", GTK_SIGNAL_FUNC(quit_rtdsp_cancel), NULL);
	gtk_signal_connect(GTK_OBJECT(quit_window), "destroy", GTK_SIGNAL_FUNC(quit_rtdsp_cancel), NULL);

	gtk_box_pack_start(GTK_BOX(button_box), button_ok, TRUE, TRUE, 0);
	gtk_box_pack_start(GTK_BOX(button_box), button_cancel, TRUE, TRUE, 0);
	gtk_container_add(GTK_CONTAINER(quit_window), vbox);

	gtk_widget_show(button_ok);
	gtk_widget_show(button_cancel);
	gtk_widget_show(button_box);
	gtk_widget_show(quit_window);

	return(TRUE);
}



/* main function */
int main(int argc, char *argv[])
{
	GtkWidget *main_window;
	GtkWidget *window_table;
	GtkWidget *button_box;
	GtkWidget *button_add;
	GtkWidget *button_del;
	GtkWidget *button_conf;
	GtkWidget *button_process;
	GtkWidget *scrolled_win;
	GtkWidget *process_button_label;
	GtkItemFactory *item_factory;
	GtkAccelGroup *accel_group;
	GtkWidget *menu_bar;
	gchar *plugin_list_titles[2]={"Plugins", "Comment"};


	gtk_init(&argc, &argv);
	main_window=gtk_window_new(GTK_WINDOW_TOPLEVEL);
	gtk_container_set_border_width(GTK_CONTAINER(main_window), 10);
	gtk_window_set_title(GTK_WINDOW(main_window), "Guitar Effects Processor");
	gtk_signal_connect(GTK_OBJECT(main_window), "delete_event", GTK_SIGNAL_FUNC(quit_rtdsp), NULL);
	gtk_signal_connect(GTK_OBJECT(main_window), "destroy", GTK_SIGNAL_FUNC(gtk_main_quit), NULL);
	
	window_table=gtk_table_new(3, 2, FALSE);
	gtk_container_add(GTK_CONTAINER(main_window), window_table);
	
	accel_group=gtk_accel_group_new();
	item_factory=gtk_item_factory_new(GTK_TYPE_MENU_BAR, "<main>", accel_group);
	gtk_item_factory_create_items(item_factory, 6, menu_items, NULL);
	gtk_accel_group_attach (accel_group, GTK_OBJECT (main_window));
	menu_bar = gtk_item_factory_get_widget (item_factory, "<main>");
	gtk_widget_show(menu_bar);
	gtk_table_attach(GTK_TABLE(window_table), menu_bar, 0, 2, 0, 1, GTK_FILL | GTK_EXPAND, 0, 0, 0);


	plugin_list=gtk_clist_new_with_titles(2, plugin_list_titles);
	gtk_clist_column_titles_passive(GTK_CLIST(plugin_list));
	gtk_clist_set_shadow_type(GTK_CLIST(plugin_list), GTK_SHADOW_OUT);
	gtk_clist_set_column_width(GTK_CLIST(plugin_list), 0, 200);
	gtk_clist_set_column_width(GTK_CLIST(plugin_list), 1, 200);
	gtk_clist_set_selection_mode(GTK_CLIST(plugin_list), GTK_SELECTION_BROWSE);
	gtk_signal_connect(GTK_OBJECT(plugin_list), "select_row", GTK_SIGNAL_FUNC(select_row_plugin_list), NULL);
	
	scrolled_win=gtk_scrolled_window_new(NULL, NULL);
	gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(scrolled_win), GTK_POLICY_AUTOMATIC, GTK_POLICY_AUTOMATIC);
	gtk_table_attach_defaults(GTK_TABLE(window_table), scrolled_win, 0, 1, 2, 3);
	gtk_scrolled_window_add_with_viewport(GTK_SCROLLED_WINDOW(scrolled_win), plugin_list);	
	gtk_widget_set_usize(scrolled_win, 450, 100);

	button_box=gtk_vbox_new(FALSE, 0);
			
	button_add=gtk_button_new_with_label("Add");
	button_del=gtk_button_new_with_label("Remove");
	button_conf=gtk_button_new_with_label("Configure");
	button_process=gtk_button_new();
	process_button_label=gtk_label_new("Start");
	gtk_container_add(GTK_CONTAINER(button_process), process_button_label);
	gtk_widget_show(process_button_label);

	gtk_signal_connect(GTK_OBJECT(button_add), "clicked", GTK_SIGNAL_FUNC(add_plugin_to_list), plugin_list);
	gtk_signal_connect(GTK_OBJECT(button_del), "clicked", GTK_SIGNAL_FUNC(remove_plugin_from_list), plugin_list);
	gtk_signal_connect(GTK_OBJECT(button_conf), "clicked", GTK_SIGNAL_FUNC(configure_plugin), NULL);
	gtk_signal_connect(GTK_OBJECT(button_process), "clicked", GTK_SIGNAL_FUNC(process_start_stop), process_button_label);

	gtk_widget_show(button_add);
	gtk_widget_show(button_del);
	gtk_widget_show(button_conf);
	gtk_widget_show(button_process);

	gtk_box_pack_start(GTK_BOX(button_box),button_add, FALSE, FALSE, 0);
	gtk_box_pack_start(GTK_BOX(button_box),button_del, FALSE, FALSE, 0);
	gtk_box_pack_start(GTK_BOX(button_box),button_conf, FALSE, FALSE, 0);
	gtk_box_pack_start(GTK_BOX(button_box),button_process, FALSE, FALSE, 0);

	gtk_table_attach(GTK_TABLE(window_table), button_box, 1, 2, 2, 3, 0, 0, 0, 0);
	gtk_widget_show(button_box);

	gtk_widget_show(plugin_list);
	gtk_widget_show(scrolled_win);
	gtk_widget_show(window_table);
	gtk_widget_show(main_window);

	gtk_main();

	return (0);
}
