#include <gtk/gtk.h>
#include "plugin.h"
#include <math.h>


#define FLT_LOSHELF 1
#define FLT_HISHELF 2
#define FLT_PEAK 3

struct filter_data_str
{
	float a0, a1, a2, b1, b2;
	float x1, x2, y1, y2;
	float cutfreq;
	float Q;
	float gain;
};


struct plugin_data
{
	GtkWidget *window;
	int sampling_frequency;
	
	struct filter_data_str peakl, peakr;
};


void *init(void)
{
	struct plugin_data *temp;

	temp=(struct plugin_data *)calloc(sizeof(struct plugin_data), 1);
	temp->peakl.gain=0.0;
	temp->peakl.cutfreq=1000.0;
	temp->peakl.Q=1.0;
	temp->peakr.gain=0.0;
	temp->peakr.cutfreq=1000.0;
	temp->peakr.Q=1.0;
	
	temp->sampling_frequency=50000;
	return temp;
}


void cleanup(void *data)
{
	struct plugin_data *temp;
	temp=data;
	if(temp->window != NULL) gtk_widget_destroy(temp->window);
	temp->window=NULL;
	free(data);
}


void destroy_window(GtkWidget *wind, struct plugin_data *data)
{
	gtk_widget_destroy(wind);
	data->window=NULL;
}




void calc_coeff_flt(int type, float smpfreq, struct filter_data_str *flt)
{
	float A, omega, sn, cs, alpha, beta;
	float norm;
	float cutfreq, gain, Q;
	
	cutfreq=flt->cutfreq;
	gain=flt->gain;
	Q=flt->Q;

	A=pow(10.0, gain / 40.0);
	omega=2.0 * M_PI * cutfreq / smpfreq;
	sn=sin(omega);
	cs=cos(omega);
	alpha=sn / (2.0 * Q);
	beta=sqrt((A * A + 1.0) / Q - ((A - 1.0) * (A - 1.0)));

	switch(type)
	{
		case FLT_HISHELF:
			norm=(A + 1.0) - (A - 1.0) * cs + beta * sn;
			flt->a0=(A * ((A + 1.0) + (A - 1.0) * cs + beta * sn)) / norm;
			flt->a1=(-2.0 * A * ((A - 1.0) + (A + 1.0) * cs)) / norm;
			flt->a2=(A * ((A + 1.0) + (A - 1.0) * cs - beta * sn)) / norm;
			flt->b1=(2.0 * ((A - 1.0) - (A + 1.0) * cs)) / norm;
			flt->b2=((A + 1.0) - (A - 1.0) * cs - beta * sn) / norm;			
			break;
		case FLT_LOSHELF:
			norm=(A + 1.0) + (A - 1.0) * cs + beta * sn;
			flt->a0=(A * ((A + 1.0) - (A - 1.0) * cs + beta * sn)) / norm;
			flt->a1=(2.0 * A * ((A - 1.0) - (A + 1.0) * cs)) / norm;
			flt->a2=(A * ((A + 1.0) - (A - 1.0) * cs - beta * sn)) / norm;
			flt->b1=(-2.0 * ((A - 1.0) + (A + 1.0) * cs)) / norm;
			flt->b2=((A + 1.0) + (A - 1.0) * cs - beta * sn) / norm;
			break;
			
		case FLT_PEAK:
			norm=1.0 + alpha / A;
			flt->a0=(1.0 + alpha * A) / norm;
			flt->a1=(-2.0 * cs) / norm;
			flt->a2=(1.0 - alpha * A) / norm;
			flt->b1=(-2.0 * cs) / norm;
			flt->b2=(1.0 - alpha / A) / norm;
			break;
	}
}




void update_adj_gain(GtkWidget *adj, struct plugin_data *pdata)
{
	pdata->peakl.gain=(GTK_ADJUSTMENT(adj))->value;
	calc_coeff_flt(FLT_PEAK, (float)pdata->sampling_frequency, &(pdata->peakl));
	pdata->peakr.gain=(GTK_ADJUSTMENT(adj))->value;
	calc_coeff_flt(FLT_PEAK, (float)pdata->sampling_frequency, &(pdata->peakr));
}


void update_adj_Q(GtkWidget *adj, struct plugin_data *pdata)
{
	pdata->peakl.Q=(GTK_ADJUSTMENT(adj))->value;
	calc_coeff_flt(FLT_PEAK, (float)pdata->sampling_frequency, &(pdata->peakl));
	pdata->peakr.Q=(GTK_ADJUSTMENT(adj))->value;
	calc_coeff_flt(FLT_PEAK, (float)pdata->sampling_frequency, &(pdata->peakr));
}

void update_adj_cutfreq(GtkWidget *adj, struct plugin_data *pdata)
{
	pdata->peakl.cutfreq=(GTK_ADJUSTMENT(adj))->value;
	calc_coeff_flt(FLT_PEAK, (float)pdata->sampling_frequency, &(pdata->peakl));
	pdata->peakr.cutfreq=(GTK_ADJUSTMENT(adj))->value;
	calc_coeff_flt(FLT_PEAK, (float)pdata->sampling_frequency, &(pdata->peakr));
}



void configure(void *data)
{
	struct plugin_data *pdata;
	GtkWidget *label, *bar, *hbox, *vbox;
	GtkObject *adj;

	pdata=data;

	if(pdata->window != NULL) return;
	pdata->window=gtk_window_new(GTK_WINDOW_TOPLEVEL);
	gtk_widget_set_usize(pdata->window, 200, 200);
	gtk_window_set_title(GTK_WINDOW(pdata->window), "1-Band Equalizer");
	gtk_container_set_border_width(GTK_CONTAINER(pdata->window), 10);
	gtk_signal_connect(GTK_OBJECT(pdata->window), "destroy", GTK_SIGNAL_FUNC(destroy_window), data);

	hbox=gtk_hbox_new(FALSE, 0);
	gtk_container_add(GTK_CONTAINER(pdata->window), hbox);
	gtk_widget_show(hbox);
	
	vbox=gtk_vbox_new(FALSE, 0);
	gtk_box_pack_start(GTK_BOX(hbox), vbox, FALSE, FALSE, 0);
	gtk_widget_show(vbox);
	label=gtk_label_new("Gain (dB)");
	gtk_box_pack_start(GTK_BOX(vbox), label, TRUE, TRUE, 0);
	gtk_widget_show(label);
	label=gtk_label_new("Freq (Hz)");
	gtk_box_pack_start(GTK_BOX(vbox), label, TRUE, TRUE, 0);
	gtk_widget_show(label);
	label=gtk_label_new("Q");
	gtk_box_pack_start(GTK_BOX(vbox), label, TRUE, TRUE, 0);
	gtk_widget_show(label);
	

	vbox=gtk_vbox_new(FALSE, 0);
	gtk_box_pack_start(GTK_BOX(hbox), vbox, TRUE, TRUE, 0);
	gtk_widget_show(vbox);

	adj=gtk_adjustment_new(pdata->peakl.gain, -24.0, 24.0, 0.5, 1.0, 0.0);
	bar=gtk_hscale_new(GTK_ADJUSTMENT(adj));
	gtk_box_pack_start(GTK_BOX(vbox), bar, TRUE, TRUE, 0);
	gtk_signal_connect(GTK_OBJECT(adj), "value_changed", GTK_SIGNAL_FUNC(update_adj_gain), pdata);
	gtk_widget_show(bar);

	adj=gtk_adjustment_new(pdata->peakl.cutfreq, 100.0, 5000.0, 1.0, 10.0, 0.0);
	bar=gtk_hscale_new(GTK_ADJUSTMENT(adj));
	gtk_box_pack_start(GTK_BOX(vbox), bar, TRUE, TRUE, 0);
	gtk_signal_connect(GTK_OBJECT(adj), "value_changed", GTK_SIGNAL_FUNC(update_adj_cutfreq), pdata);
	gtk_widget_show(bar);

	adj=gtk_adjustment_new(pdata->peakl.Q, 0.1, 4.0, 0.1, 1.0, 0.0);
	bar=gtk_hscale_new(GTK_ADJUSTMENT(adj));
	gtk_box_pack_start(GTK_BOX(vbox), bar, TRUE, TRUE, 0);
	gtk_signal_connect(GTK_OBJECT(adj), "value_changed", GTK_SIGNAL_FUNC(update_adj_Q), pdata);
	gtk_widget_show(bar);


	gtk_widget_show(pdata->window);
}



inline float filter(struct filter_data_str *fltdata, float smp)
{
	float result;

	result=(fltdata->a0 * smp) + (fltdata->a1 * fltdata->x1) + (fltdata->a2 * fltdata->x2);
	result=result - (fltdata->b1 * fltdata->y1) - (fltdata->b2 * fltdata->y2);
	fltdata->x2=fltdata->x1;
	fltdata->x1=smp;
	fltdata->y2=fltdata->y1;
	fltdata->y1=result;
	return result;
}




int process(void *data, short int *smpdata, int size, int smpfreq, int nchan)
{
	struct plugin_data *pdata;
	int i;
	float tmpsmp;

	pdata=data;

	if(((float)smpfreq * 0.5 - 500.0) <= (pdata->peakl).cutfreq) return;
	if(pdata->sampling_frequency != smpfreq)
	{
		pdata->sampling_frequency=smpfreq;
		calc_coeff_flt(FLT_PEAK, (float)smpfreq, &(pdata->peakl));
		calc_coeff_flt(FLT_PEAK, (float)smpfreq, &(pdata->peakr));		
	}

	if((nchan == 1) && (pdata->peakl.gain != 0.0))
	{
	    for(i=0; i < size; i++)
	    {
	        tmpsmp=(float)smpdata[i];

	        tmpsmp=filter(&(pdata->peakl), tmpsmp);

		if(tmpsmp > 32700.0) tmpsmp=32700.0;
		else if(tmpsmp < -32700.0) tmpsmp=-32700.0;

		smpdata[i]=(short int)tmpsmp;
	    }
	}
	else if((nchan == 2) && (pdata->peakl.gain != 0.0))
	    {
	        for(i=0; i < size; i+=2)
		{
		    tmpsmp=(float)smpdata[i];

		    tmpsmp=filter(&(pdata->peakl), tmpsmp);

		    if(tmpsmp > 32700.0) tmpsmp=32700.0;
		    else if(tmpsmp < -32700.0) tmpsmp=-32700.0;

		    smpdata[i]=(short int)tmpsmp;
		    
		    
		    tmpsmp=(float)smpdata[i + 1];

		    tmpsmp=filter(&(pdata->peakr), tmpsmp);

		    if(tmpsmp > 32700.0) tmpsmp=32700.0;
		    else if(tmpsmp < -32700.0) tmpsmp=-32700.0;

		    smpdata[i + 1]=(short int)tmpsmp;
		}
	    }

	return size;
}




Plugin plugin_info={"1-Band Equalizer",
					cleanup,
					init,
					configure,
					process
				   };


Plugin *get_plugin_info(void)
{
	return &plugin_info;
}