#include <gtk/gtk.h>
#include "plugin.h"
#include <math.h>


#define FLT_LOSHELF 1
#define FLT_HISHELF 2
#define FLT_PEAK 3
#define NUM_BANDS 13
/* 1.6 * */
/*float cut_freq_band[NUM_BANDS]={100.0, 160.0, 256.0, 409.6, 655.0, 1048.0,
				1677.0, 2684.0, 4295.0, 6872.0, 10995.0, 17592.0};*/
				
/* x * */
/*float cut_freq_band[NUM_BANDS]={20.0, 82.0, 125.0, 250.0, 500.0, 1000.0,
				2000.0, 4000.0, 8000.0};		
float band_q[NUM_BANDS]={4.0, 4.0, 4.0, 2.5, 2.5, 2.5, 2.5, 2.5, 2.5};	*/			

				
/* 1.4 * */
float cut_freq_band[NUM_BANDS]={100.0, 140.0, 196.0, 274.4, 384.16, 537.824,
				752.9, 1054.1, 1475.8, 2066.1, 2892.5, 4049.6, 5668.0};
float band_q[NUM_BANDS]={2.5, 3.5, 2.7, 3.3, 2.8, 3.16, 2.9, 3.1, 2.95, 3.05, 2.97, 3.02, 2.5};
						
int sampling_frequency=50000;

struct filter_data_str
{
	float a0, a1, a2, b1, b2;
	float x1, x2, y1, y2;
	float cutfreq;
};


struct plugin_data
{
	GtkWidget *window;
	float bandgain[NUM_BANDS];
	
	struct filter_data_str bandfilter[NUM_BANDS];
};


void *init(void)
{
	struct plugin_data *temp;
	int i;

	temp=(struct plugin_data *)calloc(sizeof(struct plugin_data), 1);
	for(i=0; i < NUM_BANDS; i++)
	{
	    temp->bandfilter[i].cutfreq=cut_freq_band[i];
	}
	return temp;
}


void cleanup(void *data)
{
	struct plugin_data *temp;
	temp=data;
	if(temp->window != NULL) gtk_widget_destroy(temp->window);
	temp->window=NULL;
	free(data);
}


void destroy_window(GtkWidget *wind, struct plugin_data *data)
{
	gtk_widget_destroy(wind);
	data->window=NULL;
}




void calc_coeff_flt(int type, float smpfreq, float cutfreq, float gain, float Q, struct filter_data_str *flt)
{
	float A, omega, sn, cs, alpha, beta;
	float norm;

	A=pow(10.0, gain / 40.0);
	omega=2.0 * M_PI * cutfreq / smpfreq;
	sn=sin(omega);
	cs=cos(omega);
	alpha=sn / (2.0 * Q);
	beta=sqrt((A * A + 1.0) / Q - ((A - 1.0) * (A - 1.0)));

	switch(type)
	{
		case FLT_HISHELF:
			norm=(A + 1.0) - (A - 1.0) * cs + beta * sn;
			flt->a0=(A * ((A + 1.0) + (A - 1.0) * cs + beta * sn)) / norm;
			flt->a1=(-2.0 * A * ((A - 1.0) + (A + 1.0) * cs)) / norm;
			flt->a2=(A * ((A + 1.0) + (A - 1.0) * cs - beta * sn)) / norm;
			flt->b1=(2.0 * ((A - 1.0) - (A + 1.0) * cs)) / norm;
			flt->b2=((A + 1.0) - (A - 1.0) * cs - beta * sn) / norm;			
			break;
		case FLT_LOSHELF:
			norm=(A + 1.0) + (A - 1.0) * cs + beta * sn;
			flt->a0=(A * ((A + 1.0) - (A - 1.0) * cs + beta * sn)) / norm;
			flt->a1=(2.0 * A * ((A - 1.0) - (A + 1.0) * cs)) / norm;
			flt->a2=(A * ((A + 1.0) - (A - 1.0) * cs - beta * sn)) / norm;
			flt->b1=(-2.0 * ((A - 1.0) + (A + 1.0) * cs)) / norm;
			flt->b2=((A + 1.0) + (A - 1.0) * cs - beta * sn) / norm;
			break;
			
		case FLT_PEAK:
			norm=1.0 + alpha / A;
			flt->a0=(1.0 + alpha * A) / norm;
			flt->a1=(-2.0 * cs) / norm;
			flt->a2=(1.0 - alpha * A) / norm;
			flt->b1=(-2.0 * cs) / norm;
			flt->b2=(1.0 - alpha / A) / norm;
			break;
	}
}




void update_adj_low(GtkWidget *adj, struct plugin_data *pdata)
{
	pdata->bandgain[0]=(GTK_ADJUSTMENT(adj))->value;
	calc_coeff_flt(FLT_LOSHELF, (float)sampling_frequency, (pdata->bandfilter[0]).cutfreq, pdata->bandgain[0], 1.0, &(pdata->bandfilter[0]));
}


void update_adj_high(GtkWidget *adj, struct plugin_data *pdata)
{
	pdata->bandgain[NUM_BANDS - 1]=(GTK_ADJUSTMENT(adj))->value;
	calc_coeff_flt(FLT_HISHELF, (float)sampling_frequency, (pdata->bandfilter[NUM_BANDS - 1]).cutfreq, pdata->bandgain[NUM_BANDS - 1], 1.0, &(pdata->bandfilter[NUM_BANDS - 1]));
}

void update_adj_mid_gain(GtkWidget *adj, float *gaindata)
{
	*gaindata=(GTK_ADJUSTMENT(adj))->value;
}

void update_adj_mid_calc(GtkWidget *adj, struct filter_data_str *pdata)
{
    calc_coeff_flt(FLT_PEAK, (float)sampling_frequency, pdata->cutfreq, (GTK_ADJUSTMENT(adj))->value, 1.0, pdata);
}




void configure(void *data)
{
	struct plugin_data *pdata;
	GtkWidget *label, *bar, *hbox, *vbox;
	GtkObject *adj;
	int i;
	char convtext[10];

	pdata=data;

	if(pdata->window != NULL) return;
	pdata->window=gtk_window_new(GTK_WINDOW_TOPLEVEL);
	gtk_widget_set_usize(pdata->window, 400, 200);
	gtk_window_set_title(GTK_WINDOW(pdata->window), "12-Band Equalizer");
	gtk_container_set_border_width(GTK_CONTAINER(pdata->window), 10);
	gtk_signal_connect(GTK_OBJECT(pdata->window), "destroy", GTK_SIGNAL_FUNC(destroy_window), data);

	vbox=gtk_vbox_new(FALSE, 0);
	gtk_container_add(GTK_CONTAINER(pdata->window), vbox);
	gtk_widget_show(vbox);
	
	hbox=gtk_hbox_new(FALSE, 0);
	gtk_box_pack_start(GTK_BOX(vbox), hbox, FALSE, FALSE, 0);
	gtk_widget_show(hbox);
	for(i=0; i < NUM_BANDS; i++)
	{
	    sprintf(convtext, "%8.0f", (pdata->bandfilter[i]).cutfreq);
	    label=gtk_label_new(convtext);
	    gtk_box_pack_start(GTK_BOX(hbox), label, TRUE, TRUE, 0);
	    gtk_widget_show(label);
	}

	hbox=gtk_hbox_new(FALSE, 0);
	gtk_box_pack_start(GTK_BOX(vbox), hbox, TRUE, TRUE, 0);
	gtk_widget_show(hbox);

	adj=gtk_adjustment_new(pdata->bandgain[0], -16.0, 16.0, 0.5, 1.0, 0.0);
	bar=gtk_vscale_new(GTK_ADJUSTMENT(adj));
	gtk_box_pack_start(GTK_BOX(hbox), bar, TRUE, TRUE, 0);
	gtk_signal_connect(GTK_OBJECT(adj), "value_changed", GTK_SIGNAL_FUNC(update_adj_low), pdata);
	gtk_widget_show(bar);

	for(i=1; i < (NUM_BANDS - 1); i++)
	{
	    adj=gtk_adjustment_new(pdata->bandgain[i], -16.0, 16.0, 0.5, 1.0, 0.0);
	    bar=gtk_vscale_new(GTK_ADJUSTMENT(adj));
	    gtk_box_pack_start(GTK_BOX(hbox), bar, TRUE, TRUE, 0);
	    gtk_signal_connect(GTK_OBJECT(adj), "value_changed", GTK_SIGNAL_FUNC(update_adj_mid_gain), &(pdata->bandgain[i]));
	    gtk_signal_connect(GTK_OBJECT(adj), "value_changed", GTK_SIGNAL_FUNC(update_adj_mid_calc), &(pdata->bandfilter[i]));	    
	    gtk_widget_show(bar);
	}

	adj=gtk_adjustment_new(pdata->bandgain[NUM_BANDS - 1], -16.0, 16.0, 0.5, 1.0, 0.0);
	bar=gtk_vscale_new(GTK_ADJUSTMENT(adj));
	gtk_box_pack_start(GTK_BOX(hbox), bar, TRUE, TRUE, 0);
	gtk_signal_connect(GTK_OBJECT(adj), "value_changed", GTK_SIGNAL_FUNC(update_adj_high), pdata);
	gtk_widget_show(bar);
	
	gtk_widget_show(pdata->window);
}



inline float filter(struct filter_data_str *fltdata, float smp)
{
	float result;

	result=(fltdata->a0 * smp) + (fltdata->a1 * fltdata->x1) + (fltdata->a2 * fltdata->x2);
	result=result - (fltdata->b1 * fltdata->y1) - (fltdata->b2 * fltdata->y2);
	fltdata->x2=fltdata->x1;
	fltdata->x1=smp;
	fltdata->y2=fltdata->y1;
	fltdata->y1=result;
	return result;
}




int process(void *data, short int *smpdata, int size, int smpfreq)
{
	struct plugin_data *pdata;
	int i, fn;
	float tmpsmp;

	pdata=data;

	if(((float)smpfreq * 0.5) <= (pdata->bandfilter[NUM_BANDS - 1]).cutfreq) return;
	if(sampling_frequency != smpfreq)
	{
		sampling_frequency=smpfreq;
		calc_coeff_flt(FLT_LOSHELF, (float)smpfreq, pdata->bandfilter[0].cutfreq, pdata->bandgain[0], band_q[0], &(pdata->bandfilter[0]));
		calc_coeff_flt(FLT_HISHELF, (float)smpfreq, pdata->bandfilter[NUM_BANDS - 1].cutfreq, pdata->bandgain[NUM_BANDS - 1], band_q[NUM_BANDS - 1], &(pdata->bandfilter[NUM_BANDS - 1]));
		for(fn=1; fn < (NUM_BANDS - 1); fn++)
		    calc_coeff_flt(FLT_PEAK, (float)smpfreq, pdata->bandfilter[fn].cutfreq, pdata->bandgain[fn], band_q[fn], &(pdata->bandfilter[fn]));
	}

	for(i=0; i < (size >> 1); i++)
	{
		tmpsmp=(float)smpdata[i];

		for(fn=0; fn < NUM_BANDS; fn++)
		{
		    if(pdata->bandgain[fn] != 0.0) tmpsmp=filter(&(pdata->bandfilter[fn]), tmpsmp);
		}

		if(tmpsmp > 32767.0) tmpsmp=32767.0;
		else if(tmpsmp < -32768.0) tmpsmp=-32768.0;

		smpdata[i]=(short int)tmpsmp;
	}

	return size;
}




Plugin plugin_info={"12-Band Equalizer Plugin",
					cleanup,
					init,
					configure,
					process
				   };


Plugin *get_plugin_info(void)
{
	return &plugin_info;
}