#include <gtk/gtk.h>
#include "plugin.h"
#include <math.h>


#define FLT_LOSHELF 1
#define FLT_HISHELF 2
#define FLT_PEAK 3
#define CUT_FREQ_L 200.0
#define CUT_FREQ_H 1200.0
#define CUT_FREQ_M 700.0


struct filter_data_str
{
	float a0, a1, a2, b1, b2;
	float x1, x2, y1, y2;
	float cutfreq;
};


struct plugin_data
{
	GtkWidget *window;
	float lowgain, highgain, midgain;
	int sampling_frequency;
	
	struct filter_data_str loshelfl, hishelfl, peakl;
	struct filter_data_str loshelfr, hishelfr, peakr;
	float process_buffer[8192];
};


void *init(void)
{
	struct plugin_data *temp;

	temp=(struct plugin_data *)calloc(sizeof(struct plugin_data), 1);
	temp->lowgain=0.0;
	temp->highgain=0.0;
	temp->midgain=0.0;
	temp->hishelfl.cutfreq=CUT_FREQ_H;
	temp->loshelfl.cutfreq=CUT_FREQ_L;
	temp->peakl.cutfreq=CUT_FREQ_M;
	temp->hishelfr.cutfreq=CUT_FREQ_H;
	temp->loshelfr.cutfreq=CUT_FREQ_L;
	temp->peakr.cutfreq=CUT_FREQ_M;
		
	temp->sampling_frequency=50000;
	return temp;
}


void cleanup(void *data)
{
	struct plugin_data *temp;
	temp=data;
	if(temp->window != NULL) gtk_widget_destroy(temp->window);
	temp->window=NULL;
	free(data);
}


void destroy_window(GtkWidget *wind, struct plugin_data *data)
{
	gtk_widget_destroy(wind);
	data->window=NULL;
}




void calc_coeff_flt(int type, float smpfreq, float cutfreq, float gain, float Q, struct filter_data_str *flt)
{
	float A, omega, sn, cs, alpha, beta;
	float norm;

	A=pow(10.0, gain / 40.0);
	omega=2.0 * M_PI * cutfreq / smpfreq;
	sn=sin(omega);
	cs=cos(omega);
	alpha=sn / (2.0 * Q);
	beta=sqrt((A * A + 1.0) / Q - ((A - 1.0) * (A - 1.0)));

	switch(type)
	{
		case FLT_HISHELF:
			norm=(A + 1.0) - (A - 1.0) * cs + beta * sn;
			flt->a0=(A * ((A + 1.0) + (A - 1.0) * cs + beta * sn)) / norm;
			flt->a1=(-2.0 * A * ((A - 1.0) + (A + 1.0) * cs)) / norm;
			flt->a2=(A * ((A + 1.0) + (A - 1.0) * cs - beta * sn)) / norm;
			flt->b1=(2.0 * ((A - 1.0) - (A + 1.0) * cs)) / norm;
			flt->b2=((A + 1.0) - (A - 1.0) * cs - beta * sn) / norm;			
			break;
		case FLT_LOSHELF:
			norm=(A + 1.0) + (A - 1.0) * cs + beta * sn;
			flt->a0=(A * ((A + 1.0) - (A - 1.0) * cs + beta * sn)) / norm;
			flt->a1=(2.0 * A * ((A - 1.0) - (A + 1.0) * cs)) / norm;
			flt->a2=(A * ((A + 1.0) - (A - 1.0) * cs - beta * sn)) / norm;
			flt->b1=(-2.0 * ((A - 1.0) + (A + 1.0) * cs)) / norm;
			flt->b2=((A + 1.0) + (A - 1.0) * cs - beta * sn) / norm;
			break;
			
		case FLT_PEAK:
			norm=1.0 + alpha / A;
			flt->a0=(1.0 + alpha * A) / norm;
			flt->a1=(-2.0 * cs) / norm;
			flt->a2=(1.0 - alpha * A) / norm;
			flt->b1=(-2.0 * cs) / norm;
			flt->b2=(1.0 - alpha / A) / norm;
			break;
	}
}




void update_adj_low(GtkWidget *adj, struct plugin_data *pdata)
{
	pdata->lowgain=(GTK_ADJUSTMENT(adj))->value;
	calc_coeff_flt(FLT_LOSHELF, (float)pdata->sampling_frequency, (pdata->loshelfl).cutfreq, pdata->lowgain, 1.0, &(pdata->loshelfl));
	calc_coeff_flt(FLT_LOSHELF, (float)pdata->sampling_frequency, (pdata->loshelfr).cutfreq, pdata->lowgain, 1.0, &(pdata->loshelfr));	
}


void update_adj_high(GtkWidget *adj, struct plugin_data *pdata)
{
	pdata->highgain=(GTK_ADJUSTMENT(adj))->value;
	calc_coeff_flt(FLT_HISHELF, (float)pdata->sampling_frequency, (pdata->hishelfl).cutfreq, pdata->highgain, 1.0, &(pdata->hishelfl));
	calc_coeff_flt(FLT_HISHELF, (float)pdata->sampling_frequency, (pdata->hishelfr).cutfreq, pdata->highgain, 1.0, &(pdata->hishelfr));	
}

void update_adj_mid(GtkWidget *adj, struct plugin_data *pdata)
{
	pdata->midgain=(GTK_ADJUSTMENT(adj))->value;
	calc_coeff_flt(FLT_PEAK, (float)pdata->sampling_frequency, (pdata->peakl).cutfreq, pdata->midgain, 1.0, &(pdata->peakl));
	calc_coeff_flt(FLT_PEAK, (float)pdata->sampling_frequency, (pdata->peakr).cutfreq, pdata->midgain, 1.0, &(pdata->peakr));	
}




void configure(void *data)
{
	struct plugin_data *pdata;
	GtkWidget *label, *bar, *hbox, *vbox;
	GtkObject *adj;

	pdata=data;

	if(pdata->window != NULL) return;
	pdata->window=gtk_window_new(GTK_WINDOW_TOPLEVEL);
	gtk_widget_set_usize(pdata->window, 200, 200);
	gtk_window_set_title(GTK_WINDOW(pdata->window), "3-Band Equalizer");
	gtk_container_set_border_width(GTK_CONTAINER(pdata->window), 10);
	gtk_signal_connect(GTK_OBJECT(pdata->window), "destroy", GTK_SIGNAL_FUNC(destroy_window), data);

	vbox=gtk_vbox_new(FALSE, 0);
	gtk_container_add(GTK_CONTAINER(pdata->window), vbox);
	gtk_widget_show(vbox);
	
	hbox=gtk_hbox_new(FALSE, 0);
	gtk_box_pack_start(GTK_BOX(vbox), hbox, FALSE, FALSE, 0);
	gtk_widget_show(hbox);
	label=gtk_label_new("Bass (dB)");
	gtk_box_pack_start(GTK_BOX(hbox), label, TRUE, TRUE, 0);
	gtk_widget_show(label);
	label=gtk_label_new("Mid (dB)");
	gtk_box_pack_start(GTK_BOX(hbox), label, TRUE, TRUE, 0);
	gtk_widget_show(label);
	label=gtk_label_new("Treble (dB)");
	gtk_box_pack_start(GTK_BOX(hbox), label, TRUE, TRUE, 0);
	gtk_widget_show(label);

	hbox=gtk_hbox_new(FALSE, 0);
	gtk_box_pack_start(GTK_BOX(vbox), hbox, TRUE, TRUE, 0);
	gtk_widget_show(hbox);

	adj=gtk_adjustment_new(pdata->lowgain, -16.0, 16.0, -0.5, -1.0, 0.0);
	bar=gtk_vscale_new(GTK_ADJUSTMENT(adj));
	gtk_box_pack_start(GTK_BOX(hbox), bar, TRUE, TRUE, 0);
	gtk_signal_connect(GTK_OBJECT(adj), "value_changed", GTK_SIGNAL_FUNC(update_adj_low), pdata);
	gtk_widget_show(bar);

	adj=gtk_adjustment_new(pdata->midgain, -16.0, 16.0, 0.5, 1.0, 0.0);
	bar=gtk_vscale_new(GTK_ADJUSTMENT(adj));
	gtk_box_pack_start(GTK_BOX(hbox), bar, TRUE, TRUE, 0);
	gtk_signal_connect(GTK_OBJECT(adj), "value_changed", GTK_SIGNAL_FUNC(update_adj_mid), pdata);
	gtk_widget_show(bar);

	adj=gtk_adjustment_new(pdata->highgain, -16.0, 16.0, 0.5, 1.0, 0.0);
	bar=gtk_vscale_new(GTK_ADJUSTMENT(adj));
	gtk_box_pack_start(GTK_BOX(hbox), bar, TRUE, TRUE, 0);
	gtk_signal_connect(GTK_OBJECT(adj), "value_changed", GTK_SIGNAL_FUNC(update_adj_high), pdata);
	gtk_widget_show(bar);
	
	gtk_widget_show(pdata->window);
}



inline float filter(struct filter_data_str *fltdata, float smp)
{
	float result;

	result=(fltdata->a0 * smp) + (fltdata->a1 * fltdata->x1) + (fltdata->a2 * fltdata->x2);
	result=result - (fltdata->b1 * fltdata->y1) - (fltdata->b2 * fltdata->y2);
	fltdata->x2=fltdata->x1;
	fltdata->x1=smp;
	fltdata->y2=fltdata->y1;
	fltdata->y1=result;
	return result;
}




int process(void *data, short int *smpdata, int size, int smpfreq, int nchan)
{
	struct plugin_data *pdata;
	int i;
	float tmpsmp;

	pdata=data;

	if(((float)smpfreq * 0.5) <= (pdata->hishelfl).cutfreq - 100.0) return;
	if(pdata->sampling_frequency != smpfreq)
	{
		pdata->sampling_frequency=smpfreq;
		calc_coeff_flt(FLT_LOSHELF, (float)smpfreq, pdata->loshelfl.cutfreq, pdata->lowgain, 4.0, &(pdata->loshelfl));
		calc_coeff_flt(FLT_HISHELF, (float)smpfreq, pdata->hishelfl.cutfreq, pdata->highgain, 2.5, &(pdata->hishelfl));
		calc_coeff_flt(FLT_PEAK, (float)smpfreq, pdata->peakl.cutfreq, pdata->midgain, 2.0, &(pdata->peakl));
		calc_coeff_flt(FLT_LOSHELF, (float)smpfreq, pdata->loshelfr.cutfreq, pdata->lowgain, 4.0, &(pdata->loshelfr));
		calc_coeff_flt(FLT_HISHELF, (float)smpfreq, pdata->hishelfr.cutfreq, pdata->highgain, 2.5, &(pdata->hishelfr));
		calc_coeff_flt(FLT_PEAK, (float)smpfreq, pdata->peakr.cutfreq, pdata->midgain, 2.0, &(pdata->peakr));
	}

	for(i=0; i < size; i++)
	{
	    pdata->process_buffer[i]=(float)smpdata[i];
	}


	if(nchan == 1)
	{

	    if(pdata->lowgain != 0.0)
	    for(i=0; i < size; i++)
	    {
		pdata->process_buffer[i]=filter(&(pdata->loshelfl), pdata->process_buffer[i]);
	    }	


	    if(pdata->midgain != 0.0)
	    for(i=0; i < size; i++)
	    {
		pdata->process_buffer[i]=filter(&(pdata->peakl), pdata->process_buffer[i]);	
	    }

	    if(pdata->highgain != 0.0)
	    for(i=0; i < size; i++)
	    {
		pdata->process_buffer[i]=filter(&(pdata->hishelfl), pdata->process_buffer[i]);	
	    }
	}
	else
	  if(nchan == 2)
	  {
		if(pdata->lowgain != 0.0)
		for(i=0; i < size; i+=2)
		{
		    pdata->process_buffer[i]=filter(&(pdata->loshelfl), pdata->process_buffer[i]);
		    pdata->process_buffer[i+1]=filter(&(pdata->loshelfr), pdata->process_buffer[i+1]);		    
		}	


		if(pdata->midgain != 0.0)
		for(i=0; i < size; i+=2)
		{
		    pdata->process_buffer[i]=filter(&(pdata->peakl), pdata->process_buffer[i]);	
		    pdata->process_buffer[i+1]=filter(&(pdata->peakr), pdata->process_buffer[i+1]);			    
		}

		if(pdata->highgain != 0.0)
		for(i=0; i < size; i+=2)
		{
		    pdata->process_buffer[i]=filter(&(pdata->hishelfl), pdata->process_buffer[i]);	
		    pdata->process_buffer[i+1]=filter(&(pdata->hishelfr), pdata->process_buffer[i+1]);			    
		}
	  }
	
	for(i=0; i < size; i++)
	{
	    if(pdata->process_buffer[i] > 32767.0) pdata->process_buffer[i]=32767.0;
	    else if(pdata->process_buffer[i] < -32768.0) pdata->process_buffer[i]=-32768.0;

	    smpdata[i]=(short int)pdata->process_buffer[i];
	}
	
	return size;
}




Plugin plugin_info={"3-Band Equalizer Plugin",
					cleanup,
					init,
					configure,
					process
				   };


Plugin *get_plugin_info(void)
{
	return &plugin_info;
}