#include <gtk/gtk.h>
#include "plugin.h"
#include <math.h>


struct plugin_data
{
	GtkWidget *window;
	float level, threshold, master_vol;
	float lowl, lowr;
};

void *init(void)
{
	struct plugin_data *temp;
	
	temp=(struct plugin_data *)malloc(sizeof(struct plugin_data));
	temp->level=1.0;
	temp->master_vol=100.0;
	temp->threshold=-10.0;
	temp->window=NULL;
	temp->lowl=0.0;
	temp->lowr=0.0;	
	
	return temp;
}

static void cleanup(void *data)
{
	if(((struct plugin_data *)data)->window != NULL)
	    gtk_widget_destroy(((struct plugin_data *)data)->window);
	free(data);
}


static void destroy_confwindow(GtkWidget *wind, gpointer data)
{
	gtk_widget_destroy(wind);
	((struct plugin_data *)data)->window=NULL;
}


static void adjustment_update(GtkWidget *adj, float *data)
{
	*data=GTK_ADJUSTMENT(adj)->value;	
}

static void configure(void *data)
{
	GtkWidget *hscrollbar;
	GtkWidget *label, *hbox;
	GtkWidget *vbox;
	GtkObject *vol_adj;
	struct plugin_data *pdata;

	pdata=data;
	
	if(pdata->window != NULL) return;

	pdata->window=gtk_window_new(GTK_WINDOW_TOPLEVEL);
	gtk_container_set_border_width(GTK_CONTAINER(pdata->window), 10);
	gtk_window_set_title(GTK_WINDOW(pdata->window), "Compressor");
	gtk_widget_set_usize(pdata->window, 300, 120);
	gtk_signal_connect(GTK_OBJECT(pdata->window), "destroy", GTK_SIGNAL_FUNC(destroy_confwindow), pdata);
	
	hbox=gtk_hbox_new(FALSE, 0);
	gtk_container_add(GTK_CONTAINER(pdata->window), hbox);
	gtk_widget_show(hbox);

	vbox=gtk_vbox_new(FALSE, 0);
	gtk_widget_show(vbox);
	label=gtk_label_new("Compression level");
	gtk_box_pack_start(GTK_BOX(vbox), label, TRUE, TRUE, 10);
	gtk_widget_show(label);

	label=gtk_label_new("Threshold (dB)");
	gtk_box_pack_start(GTK_BOX(vbox), label, TRUE, TRUE, 10);
	gtk_widget_show(label);	

	gtk_box_pack_start(GTK_BOX(hbox), vbox, TRUE, TRUE, 0);

	vbox=gtk_vbox_new(FALSE, 0);
	gtk_widget_show(vbox);
	
	vol_adj=gtk_adjustment_new(pdata->level, 1.0, 20.0, 0.5, 1.0, 0.0);
	gtk_signal_connect(vol_adj, "value_changed", GTK_SIGNAL_FUNC(adjustment_update), &(pdata->level));
	hscrollbar=gtk_hscale_new(GTK_ADJUSTMENT(vol_adj));
	gtk_widget_show(hscrollbar);
	
	gtk_box_pack_start(GTK_BOX(vbox), hscrollbar, TRUE, TRUE, 0);
	
	vol_adj=gtk_adjustment_new(pdata->threshold, -60.0, 0.0, 1.0, 1.0, 0.0);
	gtk_signal_connect(vol_adj, "value_changed", GTK_SIGNAL_FUNC(adjustment_update), &(pdata->threshold));
	hscrollbar=gtk_hscale_new(GTK_ADJUSTMENT(vol_adj));
	gtk_widget_show(hscrollbar);
	
	gtk_box_pack_start(GTK_BOX(vbox), hscrollbar, TRUE, TRUE, 0);


	gtk_box_pack_start(GTK_BOX(hbox), vbox, TRUE, TRUE, 0);

	gtk_widget_show(pdata->window);
}


static int process(void *data, short int *smpdata, int size, int sampf, int nchan)
{
	int i;
	register float temp, znak;
	float level, volume, threshold, threshold_f;	
	float compH;
	struct plugin_data *pdata;
	float ampin;
	float ampout;

	pdata=data;
	level=1.0 / pdata->level;
	volume=pdata->master_vol / 100.0;
	threshold_f=pow(10.0, pdata->threshold / 20.0);
	threshold=threshold_f * 32767.0;
	volume=1.0 / ((1.0 - threshold_f) * level + threshold_f);	
	
	if(nchan == 1)
	{
	    for(i=0; i < size; i++)
	    {
		temp=(float)smpdata[i];
		if(temp < 0.0) ampin=-temp * 0.01 + 0.99 * pdata->lowl;
		else ampin=temp * 0.01 + 0.99 * pdata->lowl;
		pdata->lowl=ampin;
		
		if(ampin > threshold)
		{
		    compH=ampin - threshold;
		    ampout=compH * level + threshold;
		    
		    temp*=ampout / ampin;
		}

	        temp*=volume;

		if(temp > 32700.0) temp=32700.0;
		else if(temp < -32700.0) temp=-32700.0;
		smpdata[i]=(short int)temp;
	    }
	}
	else
	    if(nchan == 2)
	    {
	        for(i=0; i < size; i+=2)
		{
		    temp=(float)smpdata[i];
		    if(temp < 0.0) ampin=-temp * 0.01 + 0.99 * pdata->lowl;
		    else ampin=temp * 0.01 + 0.99 * pdata->lowl;
		    pdata->lowl=ampin;

		    if(ampin > threshold)
		    {
			compH=ampin - threshold;
			ampout=compH * level + threshold;
		    
			temp*=ampout / ampin;
		    }

	    	    temp*=volume;
    
		    if(temp > 32700.0) temp=32700.0;
		    else if(temp < -32700.0) temp=-32700.0;
		    smpdata[i]=(short int)temp;
		    
		    temp=(float)smpdata[i + 1];
		    if(temp < 0.0) ampin=-temp * 0.01 + 0.99 * pdata->lowr;
		    else ampin=temp * 0.01 + 0.99 * pdata->lowr;
		    pdata->lowr=ampin;

		    if(ampin > threshold)
		    {
			compH=ampin - threshold;
			ampout=compH * level + threshold;
		    
			temp*=ampout / ampin;
		    }

	    	    temp*=volume;
    
		    if(temp > 32700.0) temp=32700.0;
		    else if(temp < -32700.0) temp=-32700.0;
		    smpdata[i + 1]=(short int)temp;
		}
	    }
	return size;
}



Plugin distortion_plugin_data={"Compressor",
cleanup,
init,
configure,
process};

Plugin *get_plugin_info(void)
{
	return &distortion_plugin_data;
}