#include <gtk/gtk.h>
#include "plugin.h"
#include <math.h>


#define MOD_DELAY_SIZE 1000
#define MAX_SAMP_FREQ 45000
#define MAX_DELAY 1000.0

struct plugin_data
{
	GtkWidget *window;
	float c_mod, dp_max, feedback, wet, dry, rate;
	float *buffer;
	float *mod_delay;

	float mod_delay_ind;
	int buffer_ind;
};

float sine[MOD_DELAY_SIZE];
float triangle[MOD_DELAY_SIZE];
int initialized=0;

void *init(void)
{
	struct plugin_data *temp;
	int i;
		
	temp=(struct plugin_data *)malloc(sizeof(struct plugin_data));
	temp->buffer=(float *)calloc((int)((float)MAX_SAMP_FREQ * MAX_DELAY / 1000.0), sizeof(float));
	temp->window=NULL;
	temp->c_mod=0.0;
	temp->dp_max=1.0;
	temp->feedback=0.0;
	temp->dry=50.0;
	temp->wet=50.0;	
	temp->rate=10.0;
	temp->mod_delay_ind=0.0;
	temp->buffer_ind=0;
	temp->mod_delay=sine;
	if(!initialized)
	{
	    initialized=1;
	    for(i=0; i < MOD_DELAY_SIZE; i++)
	    {
			sine[i]=(1.0 + sin(2.0 * M_PI * (float)i / (float)MOD_DELAY_SIZE)) / 2.0;
	    }
	}
	return temp;
}

static void cleanup(void *data)
{
	if(((struct plugin_data *)data)->window != NULL)
	    gtk_widget_destroy(((struct plugin_data *)data)->window);
	free(((struct plugin_data *)data)->buffer);
	free(data);
}


static void destroy_confwindow(GtkWidget *wind, gpointer data)
{
	gtk_widget_destroy(wind);
	((struct plugin_data *)data)->window=NULL;
}


static void adjustment_update(GtkWidget *adj, float *data)
{
	*data=GTK_ADJUSTMENT(adj)->value;	
}

static void configure(void *data)
{
	GtkWidget *hscrollbar;
	GtkWidget *label, *hbox;
	GtkWidget *vbox;
	GtkObject *vol_adj;
	struct plugin_data *pdata;

	pdata=data;
	
	if(pdata->window != NULL) return;

	pdata->window=gtk_window_new(GTK_WINDOW_TOPLEVEL);
	gtk_container_set_border_width(GTK_CONTAINER(pdata->window), 10);
	gtk_window_set_title(GTK_WINDOW(pdata->window), "Flanger Plugin");
	gtk_widget_set_usize(pdata->window, 300, 200);
	gtk_signal_connect(GTK_OBJECT(pdata->window), "destroy", GTK_SIGNAL_FUNC(destroy_confwindow), pdata);
	
	hbox=gtk_hbox_new(FALSE, 0);
	gtk_container_add(GTK_CONTAINER(pdata->window), hbox);
	gtk_widget_show(hbox);

	vbox=gtk_vbox_new(FALSE, 0);
	gtk_widget_show(vbox);
	label=gtk_label_new("Constant Delay (ms)");
	gtk_box_pack_start(GTK_BOX(vbox), label, TRUE, TRUE, 10);
	gtk_widget_show(label);
	label=gtk_label_new("Depth (ms)");
	gtk_box_pack_start(GTK_BOX(vbox), label, TRUE, TRUE, 10);
	gtk_widget_show(label);
	label=gtk_label_new("Wet (%)");
	gtk_box_pack_start(GTK_BOX(vbox), label, TRUE, TRUE, 10);
	gtk_widget_show(label);
	label=gtk_label_new("Dry (%)");
	gtk_box_pack_start(GTK_BOX(vbox), label, TRUE, TRUE, 10);
	gtk_widget_show(label);	
	label=gtk_label_new("Feedback (%)");
	gtk_box_pack_start(GTK_BOX(vbox), label, TRUE, TRUE, 10);
	gtk_widget_show(label);
	label=gtk_label_new("Rate (Hz)");
	gtk_box_pack_start(GTK_BOX(vbox), label, TRUE, TRUE, 10);
	gtk_widget_show(label);	
	
	gtk_box_pack_start(GTK_BOX(hbox), vbox, FALSE, FALSE, 0);

	vbox=gtk_vbox_new(FALSE, 0);
	gtk_widget_show(vbox);
	
	vol_adj=gtk_adjustment_new(pdata->c_mod, 0.0, 10.0, 1.0, 1.0, 0.0);
	gtk_signal_connect(vol_adj, "value_changed", GTK_SIGNAL_FUNC(adjustment_update), &(pdata->c_mod));
	hscrollbar=gtk_hscale_new(GTK_ADJUSTMENT(vol_adj));
	gtk_widget_show(hscrollbar);
	
	gtk_box_pack_start(GTK_BOX(vbox), hscrollbar, TRUE, TRUE, 0);

	vol_adj=gtk_adjustment_new(pdata->dp_max, 1.0, 10.0, 1.0, 1.0, 0.0);
	gtk_signal_connect(vol_adj, "value_changed", GTK_SIGNAL_FUNC(adjustment_update), &(pdata->dp_max));
	hscrollbar=gtk_hscale_new(GTK_ADJUSTMENT(vol_adj));
	gtk_widget_show(hscrollbar);
	
	gtk_box_pack_start(GTK_BOX(vbox), hscrollbar, TRUE, TRUE, 0);

	vol_adj=gtk_adjustment_new(pdata->wet, 0.0, 100.0, 1.0, 1.0, 0.0);
	gtk_signal_connect(vol_adj, "value_changed", GTK_SIGNAL_FUNC(adjustment_update), &(pdata->wet));
	hscrollbar=gtk_hscale_new(GTK_ADJUSTMENT(vol_adj));
	gtk_widget_show(hscrollbar);
	
	gtk_box_pack_start(GTK_BOX(vbox), hscrollbar, TRUE, TRUE, 0);
	
	vol_adj=gtk_adjustment_new(pdata->dry, 0.0, 100.0, 1.0, 1.0, 0.0);
	gtk_signal_connect(vol_adj, "value_changed", GTK_SIGNAL_FUNC(adjustment_update), &(pdata->dry));
	hscrollbar=gtk_hscale_new(GTK_ADJUSTMENT(vol_adj));
	gtk_widget_show(hscrollbar);
	
	gtk_box_pack_start(GTK_BOX(vbox), hscrollbar, TRUE, TRUE, 0);	
	
	vol_adj=gtk_adjustment_new(pdata->feedback, -100.0, 100.0, 1.0, 1.0, 0.0);
	gtk_signal_connect(vol_adj, "value_changed", GTK_SIGNAL_FUNC(adjustment_update), &(pdata->feedback));
	hscrollbar=gtk_hscale_new(GTK_ADJUSTMENT(vol_adj));
	gtk_widget_show(hscrollbar);
	
	gtk_box_pack_start(GTK_BOX(vbox), hscrollbar, TRUE, TRUE, 0);	

	vol_adj=gtk_adjustment_new(pdata->rate, 0.01, 5.0, 0.01, 1.0, 0.0);
	gtk_signal_connect(vol_adj, "value_changed", GTK_SIGNAL_FUNC(adjustment_update), &(pdata->rate));
	hscrollbar=gtk_hscale_new(GTK_ADJUSTMENT(vol_adj));
	gtk_scale_set_digits(GTK_SCALE(hscrollbar), 2);
	gtk_widget_show(hscrollbar);
	
	gtk_box_pack_start(GTK_BOX(vbox), hscrollbar, TRUE, TRUE, 0);
	

	gtk_box_pack_start(GTK_BOX(hbox), vbox, TRUE, TRUE, 0);

	gtk_widget_show(pdata->window);
}


static int process(void *data, short int *smpdata, int size, int sampf, int nchan)
{
    int i;
    register float temp, fb_temp;
    struct plugin_data *pdata;
    float const_mod_v, depth_mod_v, fb, wet, dry, r;
    int delay, delayed_smp;
    
    pdata=data;

    const_mod_v=(pdata->c_mod) * (float)sampf * (1.0 / 1000.0);
    depth_mod_v=(pdata->dp_max) * (float)sampf * (1.0 / 1000.0);
    fb=(pdata->feedback) * (1.0 / 100.0);
    wet=(pdata->wet) * (1.0 / 100.0);
    dry=(pdata->dry) * (1.0 / 100.0);    
    r=((float)MOD_DELAY_SIZE * pdata->rate) / (float)sampf;
    
    if(nchan == 1)
    {
	for(i=0; i < size; i++)
	{
    	    temp=(float)smpdata[i];
	
	    delay=(int)(depth_mod_v * (pdata->mod_delay[(int)pdata->mod_delay_ind]) + const_mod_v);
	    pdata->mod_delay_ind+=r;
	    if(pdata->mod_delay_ind >= (float)MOD_DELAY_SIZE) pdata->mod_delay_ind-=(float)MOD_DELAY_SIZE;	
	    delayed_smp=pdata->buffer_ind - delay;
	    if(delayed_smp < 0) delayed_smp+=(MAX_SAMP_FREQ * (int)(MAX_DELAY / 1000.0));
	
	    fb_temp=pdata->buffer[delayed_smp];
	    pdata->buffer[(pdata->buffer_ind)++]=fb_temp * fb + temp;
	
		/*if(fb_temp > 32767.0) fb_temp=32767.0;
		else if(fb_temp < -32768.0) fb_temp=-32768.0;*/
	
	    if(pdata->buffer_ind >= (MAX_SAMP_FREQ * (int)(MAX_DELAY / 1000.0))) pdata->buffer_ind=0;
	    
	    temp=temp * dry + fb_temp * wet;

	    if(temp > 32767.0) temp=32767.0;
	    else if(temp < -32768.0) temp=-32768.0;

	    smpdata[i]=(short int)temp;
	}
    }
    else
	if(nchan == 2)
	{
	    for(i=0; i < size; i+=2)
	    {
    		temp=(float)smpdata[i] + (float)smpdata[i + 1];
	
		delay=(int)(depth_mod_v * (pdata->mod_delay[(int)pdata->mod_delay_ind]) + const_mod_v);
		pdata->mod_delay_ind+=r;
		if(pdata->mod_delay_ind >= (float)MOD_DELAY_SIZE) pdata->mod_delay_ind-=(float)MOD_DELAY_SIZE;	
		delayed_smp=pdata->buffer_ind - delay;
		if(delayed_smp < 0) delayed_smp+=(MAX_SAMP_FREQ * (int)(MAX_DELAY / 1000.0));
	
		fb_temp=pdata->buffer[delayed_smp];
		pdata->buffer[(pdata->buffer_ind)++]=fb_temp * fb + temp;
	
		/*if(fb_temp > 32767.0) fb_temp=32767.0;
		else if(fb_temp < -32768.0) fb_temp=-32768.0;*/
	
		if(pdata->buffer_ind >= (MAX_SAMP_FREQ * (int)(MAX_DELAY / 1000.0))) pdata->buffer_ind=0;
	    
		temp=temp * dry + fb_temp * wet;

		if(temp > 32767.0) temp=32767.0;
		else if(temp < -32768.0) temp=-32768.0;

		smpdata[i]=(short int)temp;
		smpdata[i + 1]=(short int)temp;
	    }
	}
    
    return size;
}



Plugin flanger_plugin_data={"Flanger Plugin",
cleanup,
init,
configure,
process};

Plugin *get_plugin_info(void)
{
	return &flanger_plugin_data;
}