#include <gtk/gtk.h>
#include "plugin.h"
#include <math.h>


struct plugin_data_str
{
	GtkWidget *window;
	float freq;
	float smp_freq;
	float a0l,bl[4];
	float ybl[4];
	
	float a0r,br[4];
	float ybr[4];
};

void calc_coeff(struct plugin_data_str *pdata)
{
    float x;
    
    x=pow(2.718281828, -14.445 * pdata->freq / pdata->smp_freq);
    pdata->a0l=pow(1.0 - x, 4.0);
    pdata->bl[0]=4.0 * x;
    pdata->bl[1]=-6.0 * (x * x);
    pdata->bl[2]=4.0 * (x * x * x);
    pdata->bl[3]=-(x * x * x* x);
    
    pdata->a0r=pdata->a0l;
    pdata->br[0]=pdata->bl[0];
    pdata->br[1]=pdata->bl[1];    
    pdata->br[2]=pdata->bl[2];    
    pdata->br[3]=pdata->bl[3];    
}



void *init(void)
{
	struct plugin_data_str *temp;
	
	temp=(struct plugin_data_str *)malloc(sizeof(struct plugin_data_str));
	temp->freq=1000.0;
	temp->smp_freq=44100.0;
	temp->window=NULL;
	temp->ybl[0]=0.0;
	temp->ybl[1]=0.0;	
	temp->ybl[2]=0.0;	
	temp->ybl[3]=0.0;	
	temp->ybr[0]=0.0;
	temp->ybr[1]=0.0;	
	temp->ybr[2]=0.0;	
	temp->ybr[3]=0.0;	
	
	calc_coeff(temp);
	
	return temp;
}

static void cleanup(void *data)
{
	if(((struct plugin_data_str *)data)->window != NULL)
	    gtk_widget_destroy(((struct plugin_data_str *)data)->window);
	free(data);
}


static void destroy_confwindow(GtkWidget *wind, gpointer data)
{
	gtk_widget_destroy(wind);
	((struct plugin_data_str *)data)->window=NULL;
}


static void adjustment_update(GtkWidget *adj, gpointer data)
{
    struct plugin_data_str *pdata=data;
    
    pdata->freq=GTK_ADJUSTMENT(adj)->value;
    calc_coeff(pdata);
}

static void configure(void *data)
{
	GtkWidget *hscrollbar;
	GtkWidget *label, *hbox;
	GtkObject *vol_adj;
	
	if(((struct plugin_data_str *)data)->window != NULL) return;
	vol_adj=gtk_adjustment_new(((struct plugin_data_str *)data)->freq, 50.0, 8000.0, 1.0, 10.0, 0.0);
	gtk_signal_connect(GTK_OBJECT (vol_adj), "value_changed",GTK_SIGNAL_FUNC (adjustment_update), data);


	((struct plugin_data_str *)data)->window=gtk_window_new(GTK_WINDOW_TOPLEVEL);
	gtk_window_set_title(GTK_WINDOW((((struct plugin_data_str *)data)->window)), "4-Stage Lowpass Filter");
	gtk_container_set_border_width(GTK_CONTAINER((((struct plugin_data_str *)data)->window)), 10);
	/*gtk_widget_set_usize(((struct plugin_data_str *)data)->window, 300, 50);*/

	hbox=gtk_hbox_new(FALSE, 0);
	gtk_widget_show(hbox);

	label=gtk_label_new("Freq (Hz)");
	gtk_box_pack_start(GTK_BOX(hbox), label, FALSE, FALSE, 10);
	gtk_widget_show(label);

	hscrollbar=gtk_hscale_new(GTK_ADJUSTMENT(vol_adj));
	gtk_widget_set_usize(hscrollbar, 100, 50);
	gtk_scale_set_draw_value(GTK_SCALE(hscrollbar), TRUE);
	gtk_scale_set_digits(GTK_SCALE(hscrollbar), 0);
	gtk_box_pack_start(GTK_BOX(hbox), hscrollbar, TRUE, TRUE, 0);
	
	gtk_container_add(GTK_CONTAINER((((struct plugin_data_str *)data)->window)), hbox);
	gtk_signal_connect(GTK_OBJECT((((struct plugin_data_str *)data)->window)), "destroy", GTK_SIGNAL_FUNC(destroy_confwindow), data);

	gtk_widget_show(hscrollbar);
	gtk_widget_show((((struct plugin_data_str *)data)->window));
}


static int process(void *data, short int *smpdata, int size, int sampf, int nchan)
{
	int i;
	float temp;
	float newsmp;
	struct plugin_data_str *pdata=data;

	if(sampf != (int)pdata->smp_freq) 
	{
	    pdata->smp_freq=(float)sampf;
	    calc_coeff(pdata);
	}
	
	if(pdata->freq < (pdata->smp_freq / 2.0 - 100.0))
	{
	    if(nchan == 1)
	    {
		for(i=0; i < size; i++)
		{
		    temp=(float)smpdata[i];
            	    temp=temp * pdata->a0l + pdata->ybl[0] * pdata->bl[0] + 
					  pdata->ybl[1] * pdata->bl[1] +
					  pdata->ybl[2] * pdata->bl[2] +
					  pdata->ybl[3] * pdata->bl[3];
		    pdata->ybl[3]=pdata->ybl[2];
		    pdata->ybl[2]=pdata->ybl[1];
		    pdata->ybl[1]=pdata->ybl[0];
		    pdata->ybl[0]=temp;
		
		    if(temp > 32767.0) temp=32767.0;
		    else if(temp < -32768.0) temp=-32768.0;
		    smpdata[i]=(short int)temp;
		}
	    }
	    else
		if(nchan == 2)
		{
		    for(i=0; i < size; i+=2)
		    {
			temp=(float)smpdata[i];
            		temp=temp * pdata->a0l + pdata->ybl[0] * pdata->bl[0] + 
					  pdata->ybl[1] * pdata->bl[1] +
					  pdata->ybl[2] * pdata->bl[2] +
					  pdata->ybl[3] * pdata->bl[3];
			pdata->ybl[3]=pdata->ybl[2];
			pdata->ybl[2]=pdata->ybl[1];
			pdata->ybl[1]=pdata->ybl[0];
			pdata->ybl[0]=temp;
		
			if(temp > 32767.0) temp=32767.0;
			else if(temp < -32768.0) temp=-32768.0;
			smpdata[i]=(short int)temp;			
			
			temp=(float)smpdata[i + 1];
            		temp=temp * pdata->a0r + pdata->ybr[0] * pdata->br[0] + 
					  pdata->ybr[1] * pdata->br[1] +
					  pdata->ybr[2] * pdata->br[2] +
					  pdata->ybr[3] * pdata->br[3];
			pdata->ybr[3]=pdata->ybr[2];
			pdata->ybr[2]=pdata->ybr[1];
			pdata->ybr[1]=pdata->ybr[0];
			pdata->ybr[0]=temp;
		
			if(temp > 32767.0) temp=32767.0;
			else if(temp < -32768.0) temp=-32768.0;			

    			smpdata[i + 1]=(short int)temp;
		    }
		}
	}

	return size;
}



Plugin volume_plugin_data={"4-Stage Lowpass Filter",
cleanup,
init,
configure,
process};

Plugin *get_plugin_info(void)
{
	return &volume_plugin_data;
}