#include <gtk/gtk.h>
#include "plugin.h"
#include <math.h>



struct plugin_data_str
{
	GtkWidget *window;
	float threshold;
	float hold_time;
	float release_time; 
	float attack_time;
	float filter_buf;
	float volume;
	int hold_time_counter;
	int release_time_counter;
	int attack_time_counter;
	int active;
	int attacking;
	int releasing;
	float attack, release;
};

void *init(void)
{
	struct plugin_data_str *temp;
	
	temp=(struct plugin_data_str *)malloc(sizeof(struct plugin_data_str));
	temp->threshold=0.0;
	temp->hold_time=0.5;
	temp->release_time=0.5;	
	temp->attack_time=0.5;
	temp->hold_time_counter=0;
	temp->attack_time_counter=0;
	temp->attacking=0;
	temp->volume=1.0;
	temp->filter_buf=0.0;
	temp->window=NULL;
	temp->active=0;
	
	return temp;
}

static void cleanup(void *data)
{
	if(((struct plugin_data_str *)data)->window != NULL)
	    gtk_widget_destroy(((struct plugin_data_str *)data)->window);
	free(data);
}


static void destroy_confwindow(GtkWidget *wind, gpointer data)
{
	gtk_widget_destroy(wind);
	((struct plugin_data_str *)data)->window=NULL;
}


static void adjustment_update(GtkWidget *adj, float *data)
{
	*data=GTK_ADJUSTMENT(adj)->value;	
}

static void configure(void *data)
{
	GtkWidget *hscrollbar;
	GtkObject *vol_adj;
	GtkWidget *hbox;
	GtkWidget *vbox;
	GtkWidget *label;
	struct plugin_data_str *pdata=data;
	
	if(pdata->window != NULL) return;

	pdata->window=gtk_window_new(GTK_WINDOW_TOPLEVEL);
	gtk_window_set_title(GTK_WINDOW(pdata->window), "Noise Gate Plugin");
	gtk_container_set_border_width(GTK_CONTAINER(pdata->window), 10);
	gtk_widget_set_usize(pdata->window, 300, 200);
	
	vbox=gtk_vbox_new(FALSE, 0);
	gtk_container_add(GTK_CONTAINER(pdata->window), vbox);

	hbox=gtk_hbox_new(FALSE, 0);
	gtk_widget_show(hbox);

	label=gtk_label_new("Threshold (%)");
	gtk_widget_show(label);
	gtk_box_pack_start(GTK_BOX(hbox), label, FALSE, FALSE, 10);

	vol_adj=gtk_adjustment_new(pdata->threshold, 0.0, 10.0, 0.1, 0.1, 0.0);
	gtk_signal_connect(GTK_OBJECT(vol_adj), "value_changed", GTK_SIGNAL_FUNC (adjustment_update), &(pdata->threshold));

	hscrollbar=gtk_hscale_new(GTK_ADJUSTMENT(vol_adj));
	gtk_scale_set_draw_value(GTK_SCALE(hscrollbar), TRUE);
	gtk_scale_set_digits(GTK_SCALE(hscrollbar), 2);
	gtk_box_pack_start(GTK_BOX(hbox), hscrollbar, TRUE, TRUE, 0);
	gtk_box_pack_start(GTK_BOX(vbox), hbox, TRUE, TRUE, 0);
	gtk_widget_show(hscrollbar);	
	
	
	
	hbox=gtk_hbox_new(FALSE, 0);
	gtk_widget_show(hbox);

	label=gtk_label_new("Hold time (s)");
	gtk_widget_show(label);
	gtk_box_pack_start(GTK_BOX(hbox), label, FALSE, FALSE, 10);

	vol_adj=gtk_adjustment_new(pdata->hold_time, 0.001, 1.0, 0.001, 0.001, 0.0);
	gtk_signal_connect(GTK_OBJECT(vol_adj), "value_changed",GTK_SIGNAL_FUNC (adjustment_update), &(pdata->hold_time));

	hscrollbar=gtk_hscale_new(GTK_ADJUSTMENT(vol_adj));
	gtk_scale_set_draw_value(GTK_SCALE(hscrollbar), TRUE);
	gtk_scale_set_digits(GTK_SCALE(hscrollbar), 3);
	gtk_box_pack_start(GTK_BOX(hbox), hscrollbar, TRUE, TRUE, 0);
	gtk_box_pack_start(GTK_BOX(vbox), hbox, TRUE, TRUE, 0);	
	gtk_widget_show(hscrollbar);	
	
	
	hbox=gtk_hbox_new(FALSE, 0);
	gtk_widget_show(hbox);
	
	label=gtk_label_new("Release time (s)");
	gtk_widget_show(label);
	gtk_box_pack_start(GTK_BOX(hbox), label, FALSE, FALSE, 10);

	vol_adj=gtk_adjustment_new(pdata->release_time, 0.001, 1.0, 0.001, 0.001, 0.0);
	gtk_signal_connect(GTK_OBJECT(vol_adj), "value_changed",GTK_SIGNAL_FUNC (adjustment_update), &(pdata->release_time));

	hscrollbar=gtk_hscale_new(GTK_ADJUSTMENT(vol_adj));
	gtk_scale_set_draw_value(GTK_SCALE(hscrollbar), TRUE);
	gtk_scale_set_digits(GTK_SCALE(hscrollbar), 3);
	gtk_box_pack_start(GTK_BOX(hbox), hscrollbar, TRUE, TRUE, 0);
	gtk_box_pack_start(GTK_BOX(vbox), hbox, TRUE, TRUE, 0);	
	gtk_widget_show(hscrollbar);		


	hbox=gtk_hbox_new(FALSE, 0);
	gtk_widget_show(hbox);
	
	label=gtk_label_new("Attack time (s)");
	gtk_widget_show(label);
	gtk_box_pack_start(GTK_BOX(hbox), label, FALSE, FALSE, 10);

	vol_adj=gtk_adjustment_new(pdata->attack_time, 0.001, 1.0, 0.001, 0.001, 0.0);
	gtk_signal_connect(GTK_OBJECT(vol_adj), "value_changed",GTK_SIGNAL_FUNC (adjustment_update), &(pdata->attack_time));

	hscrollbar=gtk_hscale_new(GTK_ADJUSTMENT(vol_adj));
	gtk_scale_set_draw_value(GTK_SCALE(hscrollbar), TRUE);
	gtk_scale_set_digits(GTK_SCALE(hscrollbar), 3);
	gtk_box_pack_start(GTK_BOX(hbox), hscrollbar, TRUE, TRUE, 0);
	gtk_box_pack_start(GTK_BOX(vbox), hbox, TRUE, TRUE, 0);	
	gtk_widget_show(hscrollbar);		



	gtk_signal_connect(GTK_OBJECT(pdata->window), "destroy", GTK_SIGNAL_FUNC(destroy_confwindow), data);

	gtk_widget_show(vbox);
	gtk_widget_show(pdata->window);
}


static int process(void *data, short int *smpdata, int size, int sampf, int nchan)
{
	int i;
	float threshold;
	struct plugin_data_str *pdata=data;
	float amp;
	float smp, smp1, smp2;
	
	
	threshold=pdata->threshold / 100.0 * 32767.0;
	if(nchan == 1)
	{
	    for(i=0; i < size; i++)
	    {
		smp=(float)smpdata[i];
	    
		// amplitude
		if(smp >= 0.0) amp=0.01 * smp + 0.99 * pdata->filter_buf;
		    else if(smp < 0.0) amp=0.99 * pdata->filter_buf - 0.01 * smp;
		pdata->filter_buf=amp;
	    
		if(amp < threshold)
		{
		    if(!(pdata->active))
		    {
			pdata->hold_time_counter++;
			if((float)pdata->hold_time_counter > (pdata->hold_time * (float)sampf))
			{
			    pdata->active=1;
			    pdata->hold_time_counter=0;
			    pdata->release_time_counter=0;
			    pdata->attack_time_counter=0;
			    pdata->volume=1.0;
			    pdata->attacking=0;
			}
		    }
		    else
		    {
			if(pdata->attacking) 
			{
			    pdata->release_time_counter=(int)((1.0 - pdata->volume) * (pdata->release_time * (float)sampf));
			    pdata->attacking=0;
			}
			if((float)pdata->release_time_counter <= (pdata->release_time * (float)sampf)) 
			{
			    pdata->release_time_counter++;
			    pdata->volume=1.0 - ((float)pdata->release_time_counter / (pdata->release_time * (float)sampf));
			}
		    }
		}		
		else
		{
		    pdata->hold_time_counter=0;
		    if(pdata->active)
		    {
			if(!(pdata->attacking)) 
			{
			    pdata->attack_time_counter=(int)(pdata->volume * (pdata->attack_time * (float)sampf));
			    pdata->attacking=1;
			}
			if((float)pdata->attack_time_counter <= (pdata->attack_time * (float)sampf)) 
			{
			    pdata->attack_time_counter++;
			    pdata->volume=(float)pdata->attack_time_counter / (pdata->attack_time * (float)sampf);
			}
			else
			{
			    pdata->active=0;
			    pdata->attack_time_counter=0;			
			}
		    }
		}
	    
		smpdata[i]=(short int)(smp * (pdata->volume));
	    }
	}
	else if(nchan == 2)
	{
	    for(i=0; i < size; i+=2)
	    {
		smp1=(float)smpdata[i];
		smp2=(float)smpdata[i + 1];
		
		smp=smp1 * 0.5 + smp2 * 0.5;	    
		// amplitude
		if(smp >= 0.0) amp=0.01 * smp + 0.99 * pdata->filter_buf;
		    else if(smp < 0.0) amp=0.99 * pdata->filter_buf - 0.01 * smp;
		pdata->filter_buf=amp;
	    
		if(amp < threshold)
		{
		    if(!(pdata->active))
		    {
			pdata->hold_time_counter++;
			if((float)pdata->hold_time_counter > (pdata->hold_time * (float)sampf))
			{
			    pdata->active=1;
			    pdata->hold_time_counter=0;
			    pdata->release_time_counter=0;
			    pdata->attack_time_counter=0;
			    pdata->volume=1.0;
			    pdata->attacking=0;
			}
		    }
		    else
		    {
			if(pdata->attacking) 
			{
			    pdata->release_time_counter=(int)((1.0 - pdata->volume) * (pdata->release_time * (float)sampf));
			    pdata->attacking=0;
			}
			if((float)pdata->release_time_counter <= (pdata->release_time * (float)sampf)) 
			{
			    pdata->release_time_counter++;
			    pdata->volume=1.0 - ((float)pdata->release_time_counter / (pdata->release_time * (float)sampf));
			}
		    }
		}		
		else
		{
		    pdata->hold_time_counter=0;
		    if(pdata->active)
		    {
			if(!(pdata->attacking)) 
			{
			    pdata->attack_time_counter=(int)(pdata->volume * (pdata->attack_time * (float)sampf));
			    pdata->attacking=1;
			}
			if((float)pdata->attack_time_counter <= (pdata->attack_time * (float)sampf)) 
			{
			    pdata->attack_time_counter++;
			    pdata->volume=(float)pdata->attack_time_counter / (pdata->attack_time * (float)sampf);
			}
			else
			{
			    pdata->active=0;
			    pdata->attack_time_counter=0;			
			}
		    }
		}
	    
		smpdata[i]=(short int)(smp1 * (pdata->volume));
		smpdata[i + 1]=(short int)(smp2 * (pdata->volume));		
	    }
	}


    return size;
}



Plugin volume_plugin_data={"Noise Gate Plugin",
cleanup,
init,
configure,
process};

Plugin *get_plugin_info(void)
{
	return &volume_plugin_data;
}