#include <gtk/gtk.h>
#include "plugin.h"
#include <math.h>


struct plugin_data_str
{
	GtkWidget *window;
	float freq;
	float smp_freq;
	float a0,a1,b1;
	float yb;
	float xb;
};

void calc_coeff(struct plugin_data_str *pdata)
{
    float x;
    
    x=exp(-2.0 * M_PI * pdata->freq / pdata->smp_freq);
    pdata->a0=0.5 * (1.0 + x);
    pdata->a1=0.5 * -(1.0 + x);
    pdata->b1=x;
}



void *init(void)
{
	struct plugin_data_str *temp;
	
	temp=(struct plugin_data_str *)malloc(sizeof(struct plugin_data_str));
	temp->freq=1000.0;
	temp->smp_freq=44100.0;
	temp->window=NULL;
	temp->yb=0.0;
	temp->xb=0.0;
	calc_coeff(temp);

	return temp;
}

static void cleanup(void *data)
{
	if(((struct plugin_data_str *)data)->window != NULL)
	    gtk_widget_destroy(((struct plugin_data_str *)data)->window);
	free(data);
}


static void destroy_confwindow(GtkWidget *wind, gpointer data)
{
	gtk_widget_destroy(wind);
	((struct plugin_data_str *)data)->window=NULL;
}


static void adjustment_update(GtkWidget *adj, gpointer data)
{
    struct plugin_data_str *pdata=data;
    
    pdata->freq=GTK_ADJUSTMENT(adj)->value;
    calc_coeff(pdata);
}

static void configure(void *data)
{
	GtkWidget *hscrollbar;
	GtkWidget *label, *hbox;
	GtkObject *vol_adj;
	
	if(((struct plugin_data_str *)data)->window != NULL) return;
	vol_adj=gtk_adjustment_new(((struct plugin_data_str *)data)->freq, 20.0, 6000.0, 1.0, 10.0, 0.0);
	gtk_signal_connect(GTK_OBJECT (vol_adj), "value_changed",GTK_SIGNAL_FUNC (adjustment_update), data);


	((struct plugin_data_str *)data)->window=gtk_window_new(GTK_WINDOW_TOPLEVEL);
	gtk_window_set_title(GTK_WINDOW((((struct plugin_data_str *)data)->window)), "Simple Highpass Filter");
	gtk_container_set_border_width(GTK_CONTAINER((((struct plugin_data_str *)data)->window)), 10);
	/*gtk_widget_set_usize(((struct plugin_data_str *)data)->window, 300, 50);*/

	hbox=gtk_hbox_new(FALSE, 0);
	gtk_widget_show(hbox);

	label=gtk_label_new("Freq (Hz)");
	gtk_box_pack_start(GTK_BOX(hbox), label, FALSE, FALSE, 10);
	gtk_widget_show(label);

	hscrollbar=gtk_hscale_new(GTK_ADJUSTMENT(vol_adj));
	gtk_widget_set_usize(hscrollbar, 100, 50);
	gtk_scale_set_draw_value(GTK_SCALE(hscrollbar), TRUE);
	gtk_scale_set_digits(GTK_SCALE(hscrollbar), 0);
	gtk_box_pack_start(GTK_BOX(hbox), hscrollbar, TRUE, TRUE, 0);
	
	gtk_container_add(GTK_CONTAINER((((struct plugin_data_str *)data)->window)), hbox);
	gtk_signal_connect(GTK_OBJECT((((struct plugin_data_str *)data)->window)), "destroy", GTK_SIGNAL_FUNC(destroy_confwindow), data);

	gtk_widget_show(hscrollbar);
	gtk_widget_show((((struct plugin_data_str *)data)->window));
}


static int process(void *data, short int *smpdata, int size, int sampf, int nchan)
{
	int i;
	float temp;
	struct plugin_data_str *pdata=data;

	if(sampf != (int)pdata->smp_freq) 
	{
	    pdata->smp_freq=(float)sampf;
	    calc_coeff(pdata);
	}
	
	if(pdata->freq < (pdata->smp_freq / 2.0 - 100.0))
	{
	    for(i=0; i < size; i+=2)
	    {
		temp=((float)smpdata[i] + (float)smpdata[i+1]) * 0.5;
                pdata->yb=temp * pdata->a0 + pdata->xb * pdata->a1 + pdata->yb * pdata->b1;
		pdata->xb=temp;
		temp=pdata->yb;
		
		if(temp > 32767.0) temp=32767.0;
		else if(temp < -32768.0) temp=-32768.0;
		smpdata[i]=(short int)temp;
		smpdata[i+1]=(short int)temp;
	    }
	}

	return size;
}



Plugin volume_plugin_data={"Simple Highpass Filter",
cleanup,
init,
configure,
process};

Plugin *get_plugin_info(void)
{
	return &volume_plugin_data;
}