#include <gtk/gtk.h>
#include "plugin.h"
#include <math.h>


struct plugin_data_str
{
	GtkWidget *window;
	float volume;
};

void *init(void)
{
	struct plugin_data_str *temp;
	
	temp=(struct plugin_data_str *)malloc(sizeof(struct plugin_data_str));
	temp->volume=100.0;
	temp->window=NULL;
	
	return temp;
}

static void cleanup(void *data)
{
	if(((struct plugin_data_str *)data)->window != NULL)
	    gtk_widget_destroy(((struct plugin_data_str *)data)->window);
	free(data);
}


static void destroy_confwindow(GtkWidget *wind, gpointer data)
{
	gtk_widget_destroy(wind);
	((struct plugin_data_str *)data)->window=NULL;
}


static void adjustment_update(GtkWidget *adj, gpointer data)
{
	((struct plugin_data_str *)data)->volume=GTK_ADJUSTMENT(adj)->value;	
}

static void configure(void *data)
{
	GtkWidget *hscrollbar;
	GtkWidget *label, *hbox;
	GtkObject *vol_adj;
	
	if(((struct plugin_data_str *)data)->window != NULL) return;
	vol_adj=gtk_adjustment_new(((struct plugin_data_str *)data)->volume, 0.0, 500.0, 1.0, 10.0, 0.0);
	gtk_signal_connect(GTK_OBJECT (vol_adj), "value_changed",GTK_SIGNAL_FUNC (adjustment_update), data);


	((struct plugin_data_str *)data)->window=gtk_window_new(GTK_WINDOW_TOPLEVEL);
	gtk_window_set_title(GTK_WINDOW((((struct plugin_data_str *)data)->window)), "Volume Change Plugin");
	gtk_container_set_border_width(GTK_CONTAINER((((struct plugin_data_str *)data)->window)), 10);
	/*gtk_widget_set_usize(((struct plugin_data_str *)data)->window, 300, 50);*/

	hbox=gtk_hbox_new(FALSE, 0);
	gtk_widget_show(hbox);

	label=gtk_label_new("Volume (%)");
	gtk_box_pack_start(GTK_BOX(hbox), label, FALSE, FALSE, 10);
	gtk_widget_show(label);

	hscrollbar=gtk_hscale_new(GTK_ADJUSTMENT(vol_adj));
	gtk_widget_set_usize(hscrollbar, 100, 50);
	gtk_scale_set_draw_value(GTK_SCALE(hscrollbar), TRUE);
	gtk_scale_set_digits(GTK_SCALE(hscrollbar), 0);
	gtk_box_pack_start(GTK_BOX(hbox), hscrollbar, TRUE, TRUE, 0);
	
	gtk_container_add(GTK_CONTAINER((((struct plugin_data_str *)data)->window)), hbox);
	gtk_signal_connect(GTK_OBJECT((((struct plugin_data_str *)data)->window)), "destroy", GTK_SIGNAL_FUNC(destroy_confwindow), data);

	gtk_widget_show(hscrollbar);
	gtk_widget_show((((struct plugin_data_str *)data)->window));
}


static int process(void *data, short int *smpdata, int size, int sampf, int nchan)
{
	int i;
	int mul_value;
	int temp;

	mul_value=(int)((((struct plugin_data_str *)data)->volume) / 100.0 * 32768.0);
	if(nchan == 1 || nchan == 2)
	{
	    for(i=0; i < size; i++)
	    {
		temp=(int)smpdata[i];
                temp=(temp * mul_value) >> 15;
		if(temp > 32767.0) temp=32767.0;
		else if(temp < -32768.0) temp=-32768.0;
		smpdata[i]=(short int)temp;
	    }
	}
	return size;
}



Plugin volume_plugin_data={"Volume Change Plugin",
cleanup,
init,
configure,
process};

Plugin *get_plugin_info(void)
{
	return &volume_plugin_data;
}